"""
This module defines the wrapper for ML engines which abstracts away a lot of complexity.

In particular, three big components are included:

    - `BaseMLEngineExec` class: this class wraps any object that inherits from `BaseMLEngine` and exposes some endpoints
      normally associated with a DB handler (e.g. `native_query`, `get_tables`), as well as other ML-specific behaviors,
      like `learn()` or `predict()`. Note that while these still have to be implemented at the engine level, the burden
      on that class is lesser given that it only needs to return a pandas DataFrame. It's this class that will take said
      output and format it into the HandlerResponse instance that MindsDB core expects.

    - `learn_process` method: handles async dispatch of the `learn` method in an engine, as well as registering all
      models inside of the internal MindsDB registry.

    - `predict_process` method: handles async dispatch of the `predict` method in an engine.

"""

import socket
import contextlib
import datetime as dt
from types import ModuleType
from typing import Optional, Union

import pandas as pd
from sqlalchemy import func, null
from sqlalchemy.sql.functions import coalesce

from mindsdb.utilities.config import Config
import mindsdb.interfaces.storage.db as db
from mindsdb.__about__ import __version__ as mindsdb_version
from mindsdb.utilities.hooks import after_predict as after_predict_hook
from mindsdb.interfaces.model.functions import (
    get_model_record
)
from mindsdb.integrations.libs.const import PREDICTOR_STATUS
from mindsdb.interfaces.database.database import DatabaseController
from mindsdb.utilities.context import context as ctx
from mindsdb.interfaces.model.functions import get_model_records
from mindsdb.utilities.functions import mark_process
import mindsdb.utilities.profiler as profiler
from mindsdb.utilities.ml_task_queue.producer import MLTaskProducer
from mindsdb.utilities.ml_task_queue.const import ML_TASK_TYPE
from mindsdb.integrations.libs.process_cache import process_cache, empty_callback, MLProcessException

try:
    import torch.multiprocessing as mp
except Exception:
    import multiprocessing as mp
mp_ctx = mp.get_context('spawn')


class MLEngineException(Exception):
    pass


class BaseMLEngineExec:

    def __init__(self, name: str, integration_id: int, handler_module: ModuleType):
        """ML handler interface

        Args:
            name (str): name of the ml_engine
            integration_id (int): id of the ml_engine
            handler_module (ModuleType): module of the ml_engine
        """
        self.name = name
        self.config = Config()
        self.integration_id = integration_id
        self.engine = handler_module.name
        self.handler_module = handler_module

        self.database_controller = DatabaseController()

        self.base_ml_executor = process_cache
        if self.config['ml_task_queue']['type'] == 'redis':
            self.base_ml_executor = MLTaskProducer()

    @profiler.profile()
    def learn(
        self, model_name, project_name,
        data_integration_ref=None,
        fetch_data_query=None,
        problem_definition=None,
        join_learn_process=False,
        label=None,
        is_retrain=False,
        set_active=True,
    ):
        """ Trains a model given some data-gathering SQL statement. """

        # may or may not be provided (e.g. 0-shot models do not need it), so engine will handle it
        target = problem_definition.get('target', [''])  # db.Predictor expects Column(Array(String))

        project = self.database_controller.get_project(name=project_name)

        self.create_validation(target, problem_definition, self.integration_id)

        predictor_record = db.Predictor(
            company_id=ctx.company_id,
            name=model_name,
            integration_id=self.integration_id,
            data_integration_ref=data_integration_ref,
            fetch_data_query=fetch_data_query,
            mindsdb_version=mindsdb_version,
            to_predict=target,
            learn_args=problem_definition,
            data={'name': model_name},
            project_id=project.id,
            training_data_columns_count=None,
            training_data_rows_count=None,
            training_start_at=dt.datetime.now(),
            status=PREDICTOR_STATUS.GENERATING,
            label=label,
            hostname=socket.gethostname(),
            version=(
                db.session.query(
                    coalesce(func.max(db.Predictor.version), 1) + (1 if is_retrain else 0)
                ).filter_by(
                    company_id=ctx.company_id,
                    name=model_name,
                    project_id=project.id,
                    deleted_at=null()
                ).scalar_subquery()),
            active=(not is_retrain),  # if create then active
        )

        db.serializable_insert(predictor_record)

        with self._catch_exception(model_name):
            task = self.base_ml_executor.apply_async(
                task_type=ML_TASK_TYPE.LEARN,
                model_id=predictor_record.id,
                payload={
                    'handler_meta': {
                        'module_path': self.handler_module.__package__,
                        'engine': self.engine,
                        'integration_id': self.integration_id
                    },
                    'context': ctx.dump(),
                    'problem_definition': problem_definition,
                    'set_active': set_active,
                    'data_integration_ref': data_integration_ref,
                    'fetch_data_query': fetch_data_query,
                    'project_name': project_name
                }
            )

            if join_learn_process is True:
                task.result()
                predictor_record = db.Predictor.query.get(predictor_record.id)
                db.session.refresh(predictor_record)
            else:
                # to prevent memory leak need to add any callback
                task.add_done_callback(empty_callback)

        return predictor_record

    def describe(self, model_id: int, attribute: Optional[str] = None) -> pd.DataFrame:
        with self._catch_exception(model_id):
            task = self.base_ml_executor.apply_async(
                task_type=ML_TASK_TYPE.DESCRIBE,
                model_id=model_id,
                payload={
                    'handler_meta': {
                        'module_path': self.handler_module.__package__,
                        'engine': self.engine,
                        'integration_id': self.integration_id
                    },
                    'attribute': attribute,
                    'context': ctx.dump()
                }
            )
            result = task.result()
        return result

    @profiler.profile()
    @mark_process(name='predict')
    def predict(self, model_name: str, data: list, pred_format: str = 'dict',
                project_name: str = None, version=None, params: dict = None):
        """ Generates predictions with some model and input data. """
        if isinstance(data, dict):
            data = [data]
        df = pd.DataFrame(data)
        kwargs = {
            'name': model_name,
            'ml_handler_name': self.name,
            'project_name': project_name
        }
        if version is None:
            kwargs['active'] = True
        else:
            kwargs['active'] = None
            kwargs['version'] = version
        predictor_record = get_model_record(**kwargs)
        if predictor_record is None:
            if version is not None:
                model_name = f'{model_name}.{version}'
            raise Exception(f"Error: model '{model_name}' does not exists!")
        if predictor_record.status != PREDICTOR_STATUS.COMPLETE:
            raise Exception("Error: model creation not completed")

        using = {} if params is None else params
        args = {
            'pred_format': pred_format,
            'predict_params': using,
            'using': using
        }

        with self._catch_exception(model_name):
            task = self.base_ml_executor.apply_async(
                task_type=ML_TASK_TYPE.PREDICT,
                model_id=predictor_record.id,
                payload={
                    'handler_meta': {
                        'module_path': self.handler_module.__package__,
                        'engine': self.engine,
                        'integration_id': self.integration_id
                    },
                    'context': ctx.dump(),
                    'predictor_record': predictor_record,
                    'args': args
                },
                dataframe=df
            )
            predictions = task.result()

        # mdb indexes
        if '__mindsdb_row_id' not in predictions.columns and '__mindsdb_row_id' in df.columns:
            predictions['__mindsdb_row_id'] = df['__mindsdb_row_id']

        after_predict_hook(
            company_id=ctx.company_id,
            predictor_id=predictor_record.id,
            rows_in_count=df.shape[0],
            columns_in_count=df.shape[1],
            rows_out_count=len(predictions)
        )
        return predictions

    def create_validation(self, target, args, integration_id):
        with self._catch_exception():
            task = self.base_ml_executor.apply_async(
                task_type=ML_TASK_TYPE.CREATE_VALIDATION,
                model_id=0,     # can not be None
                payload={
                    'context': ctx.dump(),
                    'target': target,
                    'args': args,
                    'handler_meta': {
                        'module_path': self.handler_module.__package__,
                        'engine': self.engine,
                        'integration_id': integration_id
                    },
                }
            )
            result = task.result()
        return result

    def update(self, args: dict, model_id: int):
        with self._catch_exception(model_id):
            task = self.base_ml_executor.apply_async(
                task_type=ML_TASK_TYPE.UPDATE,
                model_id=model_id,
                payload={
                    'context': ctx.dump(),
                    'args': args,
                    'handler_meta': {
                        'module_path': self.handler_module.__package__,
                        'engine': self.engine,
                        'integration_id': self.integration_id
                    },
                }
            )
            result = task.result()
        return result

    def update_engine(self, connection_args):
        with self._catch_exception():
            task = self.base_ml_executor.apply_async(
                task_type=ML_TASK_TYPE.UPDATE_ENGINE,
                model_id=0,     # can not be None
                payload={
                    'context': ctx.dump(),
                    'connection_args': connection_args,
                    'handler_meta': {
                        'module_path': self.handler_module.__package__,
                        'engine': self.engine,
                        'integration_id': self.integration_id
                    },
                }
            )
            result = task.result()
        return result

    def create_engine(self, connection_args: dict, integration_id: int) -> None:
        with self._catch_exception():
            task = self.base_ml_executor.apply_async(
                task_type=ML_TASK_TYPE.CREATE_ENGINE,
                model_id=0,     # can not be None
                payload={
                    'context': ctx.dump(),
                    'connection_args': connection_args,
                    'handler_meta': {
                        'module_path': self.handler_module.__package__,
                        'engine': self.engine,
                        'integration_id': integration_id
                    },
                }
            )
            result = task.result()
        return result

    @profiler.profile()
    def finetune(
            self, model_name, project_name,
            base_model_version: int,
            data_integration_ref=None,
            fetch_data_query=None,
            join_learn_process=False,
            label=None,
            set_active=True,
            args: Optional[dict] = None
    ):
        # generate new record from latest version as starting point
        project = self.database_controller.get_project(name=project_name)

        search_args = {
            'active': None,
            'name': model_name,
            'status': PREDICTOR_STATUS.COMPLETE
        }
        if base_model_version is not None:
            search_args['version'] = base_model_version
        else:
            search_args['active'] = True
        predictor_records = get_model_records(**search_args)
        if len(predictor_records) == 0:
            raise Exception("Can't find suitable base model")

        predictor_records.sort(key=lambda x: x.training_stop_at, reverse=True)
        predictor_records = [x for x in predictor_records if x.training_stop_at is not None]
        base_predictor_record = predictor_records[0]

        learn_args = base_predictor_record.learn_args
        learn_args['using'] = args if not learn_args.get('using', False) else {**learn_args['using'], **args}

        self.create_validation(
            target=base_predictor_record.to_predict,
            args=learn_args,
            integration_id=self.integration_id
        )

        predictor_record = db.Predictor(
            company_id=ctx.company_id,
            name=model_name,
            integration_id=self.integration_id,
            data_integration_ref=data_integration_ref,
            fetch_data_query=fetch_data_query,
            mindsdb_version=mindsdb_version,
            to_predict=base_predictor_record.to_predict,
            learn_args=learn_args,
            data={'name': model_name},
            project_id=project.id,
            training_data_columns_count=None,
            training_data_rows_count=None,
            training_start_at=dt.datetime.now(),
            status=PREDICTOR_STATUS.GENERATING,
            label=label,
            hostname=socket.gethostname(),
            version=(
                db.session.query(
                    coalesce(func.max(db.Predictor.version), 1) + 1
                ).filter_by(
                    company_id=ctx.company_id,
                    name=model_name,
                    project_id=project.id,
                    deleted_at=null()
                ).scalar_subquery()
            ),
            active=False
        )
        db.serializable_insert(predictor_record)

        with self._catch_exception(model_name):
            task = self.base_ml_executor.apply_async(
                task_type=ML_TASK_TYPE.FINETUNE,
                model_id=predictor_record.id,
                payload={
                    'handler_meta': {
                        'module_path': self.handler_module.__package__,
                        'engine': self.engine,
                        'integration_id': self.integration_id
                    },
                    'context': ctx.dump(),
                    'model_id': predictor_record.id,
                    'problem_definition': predictor_record.learn_args,
                    'set_active': set_active,
                    'base_model_id': base_predictor_record.id,
                    'data_integration_ref': data_integration_ref,
                    'fetch_data_query': fetch_data_query,
                    'project_name': project_name
                }
            )

            if join_learn_process is True:
                task.result()
                predictor_record = db.Predictor.query.get(predictor_record.id)
                db.session.refresh(predictor_record)
            else:
                # to prevent memory leak need to add any callback
                task.add_done_callback(empty_callback)

        return predictor_record

    @contextlib.contextmanager
    def _catch_exception(self, model_identifier: Optional[Union[int, str]] = None):
        try:
            yield
        except (ImportError, ModuleNotFoundError):
            raise
        except Exception as e:
            if type(e) is MLProcessException:
                e = e.base_exception
            msg = str(e).strip()
            if msg == '':
                msg = e.__class__.__name__
            model_identifier = '' if model_identifier is None else f'/{model_identifier}'
            msg = f'[{self.name}{model_identifier}]: {msg}'
            raise MLEngineException(msg) from e



import socketserver as SocketServer
import socket
import struct
import bson
import traceback
from bson import codec_options
from collections import OrderedDict
from abc import abstractmethod
from bson.codec_options import CodecOptions
from bson.codec_options import TypeCodec
from bson.codec_options import TypeRegistry
import numpy as np
import datetime as dt

import mindsdb.api.mongo.functions as helpers
from mindsdb.api.mongo.classes import RespondersCollection, Session
from mindsdb.interfaces.storage import db
from mindsdb.interfaces.model.model_controller import ModelController
from mindsdb.interfaces.database.integrations import integration_controller
from mindsdb.interfaces.database.projects import ProjectController
from mindsdb.interfaces.database.database import DatabaseController
from mindsdb.utilities.context import context as ctx
from mindsdb.utilities import log

OP_REPLY = 1
OP_UPDATE = 2001
OP_INSERT = 2002
OP_QUERY = 2004
OP_GET_MORE = 2005
OP_DELETE = 2006
OP_KILL_CURSORS = 2007
OP_MSG = 2013

BYTE = '<b'
INT = '<i'
UINT = '<I'
LONG = '<q'

logger = log.getLogger(__name__)


class NPIntCodec(TypeCodec):
    python_type = np.int64
    bson_type = bson.int64.Int64

    def transform_python(self, value):
        return bson.int64.Int64(value)

    def transform_bson(self, value):
        return np.int(value)


class DateCodec(TypeCodec):
    python_type = dt.date
    bson_type = bson.datetime.datetime

    def transform_python(self, value):
        return dt.datetime(value.year, value.month, value.day)

    def transform_bson(self, value):
        return dt.datetime(value.year, value.month, value.day)


def fallback_encoder(value):
    return str(value)


type_registry = TypeRegistry([NPIntCodec(), DateCodec()], fallback_encoder=fallback_encoder)


def unpack(format, buffer, start=0):
    end = start + struct.calcsize(format)
    return struct.unpack(format, buffer[start:end])[0], end


def get_utf8_string(buffer, start=0):
    end = buffer.index(b"\x00", start)
    s = buffer[start:end].decode('utf8')
    return s, end + 1


CODEC_OPTIONS = codec_options.CodecOptions(document_class=OrderedDict)


def decode_documents(buffer, start, content_size):
    docs = bson.decode_all(buffer[start:start + content_size], CODEC_OPTIONS)
    return docs, start + content_size


class OperationResponder():
    def __init__(self, responders):
        self.responders = responders

    @abstractmethod
    def handle(self, query_bytes):
        pass

    @abstractmethod
    def to_bytes(self, response, request_id):
        pass


# NOTE probably, it need only for mongo version < 3.6
class OpInsertResponder(OperationResponder):
    def handle(self, buffer, request_id, mindsdb_env, session):
        flags, pos = unpack(UINT, buffer)
        namespace, pos = get_utf8_string(buffer, pos)
        query = bson.decode_all(buffer[pos:], CODEC_OPTIONS)
        responder = self.responders.find_match(query)
        assert responder is not None, 'query cant be processed'

        request_args = {
            'request_id': request_id
        }

        documents = responder.handle(query, request_args, mindsdb_env, session)

        return documents

    def to_bytes(self, response, request_id):
        pass


OP_MSG_FLAGS = {
    'checksumPresent': 0,
    'moreToCome': 1,
    'exhaustAllowed': 16
}


# NOTE used in mongo version > 3.6
class OpMsgResponder(OperationResponder):
    def handle(self, buffer, request_id, mindsdb_env, session):
        query = OrderedDict()
        flags, pos = unpack(UINT, buffer)

        checksum_present = bool(flags & (1 << OP_MSG_FLAGS['checksumPresent']))
        if checksum_present:
            msg_len = len(buffer) - 4
        else:
            msg_len = len(buffer)

        # sections
        while pos < msg_len:
            kind, pos = unpack(BYTE, buffer, pos)
            if kind == 0:
                # body
                section_size, _ = unpack(INT, buffer, pos)
                docs, pos = decode_documents(buffer, pos, section_size)
                query.update(docs[0])
            elif kind == 1:
                # Document
                section_size, pos = unpack(INT, buffer, pos)
                seq_id, pos = get_utf8_string(buffer, pos)
                docs_len = section_size - struct.calcsize(INT) - len(seq_id) - 1
                docs, pos = decode_documents(buffer, pos, docs_len)
                query[seq_id] = docs

        remaining = len(buffer) - pos
        if checksum_present:
            if remaining != 4:
                raise Exception('should be checksum at the end of message')
            # TODO read and check checksum
        elif remaining != 0:
            raise Exception('is bytes left after msg parsing')

        logger.debug(f'GET OpMSG={query}')

        responder = self.responders.find_match(query)
        assert responder is not None, 'query cant be processed'

        request_args = {
            'request_id': request_id,
            'database': query['$db']
        }

        documents = responder.handle(query, request_args, mindsdb_env, session)

        return documents

    def to_bytes(self, response, request_id, is_error=False):
        if is_error:
            flags = struct.pack("<I", 2)
        else:
            flags = struct.pack("<I", 0)  # TODO
        payload_type = struct.pack("<b", 0)  # TODO

        codec_options = CodecOptions(type_registry=type_registry)
        payload_data = bson.BSON.encode(response, codec_options=codec_options)
        data = b''.join([flags, payload_type, payload_data])

        reply_id = 0  # TODO add seq here
        response_to = request_id

        header = struct.pack("<iiii", 16 + len(data), reply_id, response_to, OP_MSG)
        return header + data


# NOTE used in any mongo shell version
class OpQueryResponder(OperationResponder):
    def handle(self, buffer, request_id, mindsdb_env, session):
        # https://docs.mongodb.com/manual/reference/mongodb-wire-protocol/#wire-op-query
        flags, pos = unpack(UINT, buffer)
        namespace, pos = get_utf8_string(buffer, pos)
        is_command = namespace.endswith('.$cmd')
        num_to_skip, pos = unpack(INT, buffer, pos)
        num_to_return, pos = unpack(INT, buffer, pos)
        docs = bson.decode_all(buffer[pos:], CODEC_OPTIONS)

        query = docs[0]  # docs = [query, returnFieldsSelector]

        logger.debug(f'GET OpQuery={query}')

        responder = self.responders.find_match(query)
        assert responder is not None, 'query cant be processed'

        request_args = {
            'num_to_skip': num_to_skip,
            'num_to_return': num_to_return,
            'request_id': request_id,
            'is_command': is_command
        }

        documents = responder.handle(query, request_args, mindsdb_env, session)

        return documents

    def to_bytes(self, request, request_id):
        flags = struct.pack("<i", 0)  # TODO
        cursor_id = struct.pack("<q", 0)  # TODO
        starting_from = struct.pack("<i", 0)  # TODO
        number_returned = struct.pack("<i", len([request]))
        reply_id = 123  # TODO
        response_to = request_id

        logger.debug(f'RET docs={request}')

        data = b''.join([flags, cursor_id, starting_from, number_returned])
        data += b''.join([bson.BSON.encode(doc) for doc in [request]])

        message = struct.pack("<i", 16 + len(data))
        message += struct.pack("<i", reply_id)
        message += struct.pack("<i", response_to)
        message += struct.pack("<i", OP_REPLY)

        return message + data


class MongoRequestHandler(SocketServer.BaseRequestHandler):
    _stopped = False

    def _init_ssl(self):
        import ssl
        import tempfile
        import atexit
        import os

        from mindsdb.utilities.wizards import make_ssl_cert

        CERT_PATH = tempfile.mkstemp(prefix='mindsdb_cert_', text=True)[1]
        make_ssl_cert(CERT_PATH)
        atexit.register(lambda: os.remove(CERT_PATH))

        ssl_context = ssl.SSLContext()
        ssl_context.load_cert_chain(CERT_PATH)
        ssl_socket = ssl_context.wrap_socket(
            self.request,
            server_side=True,
            do_handshake_on_connect=True
        )
        self.request = ssl_socket

    def handle(self):
        ctx.set_default()
        logger.debug('connect')
        logger.debug(str(self.server.socket))

        self.session = Session(self.server.mindsdb_env)

        first_byte = self.request.recv(1, socket.MSG_PEEK)
        if first_byte == b'\x16':
            # TLS 'client hello' starts from \x16
            self._init_ssl()

        while True:
            header = self._read_bytes(16)
            if header is False:
                # connection closed by client
                break
            length, pos = unpack(INT, header)
            request_id, pos = unpack(INT, header, pos)
            response_to, pos = unpack(INT, header, pos)
            opcode, pos = unpack(INT, header, pos)
            logger.debug(f'GET length={length} id={request_id} opcode={opcode}')
            msg_bytes = self._read_bytes(length - pos)
            answer = self.get_answer(request_id, opcode, msg_bytes)
            if answer is not None:
                self.request.send(answer)

            db.session.close()

    def get_answer(self, request_id, opcode, msg_bytes):
        if opcode not in self.server.operationsHandlersMap:
            raise NotImplementedError(f'Unknown opcode {opcode}')
        responder = self.server.operationsHandlersMap[opcode]
        assert responder is not None, 'error'
        try:
            response = responder.handle(msg_bytes, request_id, self.session.mindsdb_env, self.session)
            if response is None:
                return None
        except Exception as e:
            logger.error(e)
            response = {
                "ok": 0,
                "errmsg": f'{str(e)} : {traceback.format_exc()}',
                "code": 2,
                "codeName": "BadValue"
            }
            return responder.to_bytes(response, request_id, is_error=True)

        return responder.to_bytes(response, request_id)

    def _read_bytes(self, length):
        buffer = b''
        while length:
            chunk = self.request.recv(length)
            if chunk == b'':
                logger.debug('Connection closed')
                return False

            length -= len(chunk)
            buffer += chunk
        return buffer


class MongoServer(SocketServer.ThreadingMixIn, SocketServer.TCPServer):
    def __init__(self, config):
        mongodb_config = config['api'].get('mongodb')
        assert mongodb_config is not None, 'is no mongodb config!'
        host = mongodb_config['host']
        port = mongodb_config['port']
        logger.debug(f'start mongo server on {host}:{port}')

        super().__init__((host, int(port)), MongoRequestHandler)

        self.mindsdb_env = {
            'config': config,
            'model_controller': ModelController(),
            'integration_controller': integration_controller,
            'project_controller': ProjectController(),
            'database_controller': DatabaseController()
        }

        respondersCollection = RespondersCollection()

        opQueryResponder = OpQueryResponder(respondersCollection)
        opMsgResponder = OpMsgResponder(respondersCollection)
        opInsertResponder = OpInsertResponder(respondersCollection)

        self.operationsHandlersMap = {
            OP_QUERY: opQueryResponder,
            OP_MSG: opMsgResponder,
            OP_INSERT: opInsertResponder
        }

        respondersCollection.add(
            when={'drop': 'system.sessions'},
            result={'ok': 1}
        )
        respondersCollection.add(
            when={'update': 'system.version'},
            result={'ok': 1}
        )
        respondersCollection.add(
            when={'setFeatureCompatibilityVersion': helpers.is_true},
            result={'ok': 1}
        )
        # OpMSG=OrderedDict([('features', 1), ('$clusterTime', OrderedDict([('clusterTime', Timestamp(1599748325, 1)), ('signature', OrderedDict([('hash', b'\xb8\xc3\x03\x18\xca\xe6bh\xf0\xcb47,\x924\x8a >\xfc\x91'), ('keyId', 6870854312365391875)]))])), ('$configServerState', OrderedDict([('opTime', OrderedDict([('ts', Timestamp(1599748325, 1)), ('t', 1)]))])), ('$db', 'admin')])
        respondersCollection.add(
            when={'features': helpers.is_true},
            result={'ok': 1}
        )
        # OpMSG=OrderedDict([('serverStatus', 1), ('$clusterTime', OrderedDict([('clusterTime', Timestamp(1599748366, 1)), ('signature', OrderedDict([('hash', b'\xa1E}\xbbIU\xc2D\x95++\x82\x88\xb5\x84\xf5\xda)+B'), ('keyId', 6870854312365391875)]))])), ('$configServerState', OrderedDict([('opTime', OrderedDict([('ts', Timestamp(1599748366, 1)), ('t', 1)]))])), ('$db', 'admin')])
        respondersCollection.add(
            when={'serverStatus': helpers.is_true},
            result={'ok': 1}
        )
        # OpMSG=OrderedDict([('ismaster', 1), ('$db', 'admin'), ('$clusterTime', OrderedDict([('clusterTime', Timestamp(1599749031, 1)), ('signature', OrderedDict([('hash', b'6\x87\xd5Y\xa7\xc7\xcf$\xab\x1e\xa2{\xe5B\xe5\x99\xdbl\x8d\xf4'), ('keyId', 6870854312365391875)]))])), ('$client', OrderedDict([('application', OrderedDict([('name', 'MongoDB Shell')])), ('driver', OrderedDict([('name', 'MongoDB Internal Client'), ('version', '3.6.3')])), ('os', OrderedDict([('type', 'Linux'), ('name', 'Ubuntu'), ('architecture', 'x86_64'), ('version', '18.04')])), ('mongos', OrderedDict([('host', 'maxs-comp:27103'), ('client', '127.0.0.1:52148'), ('version', '3.6.3')]))])), ('$configServerState', OrderedDict([('opTime', OrderedDict([('ts', Timestamp(1599749031, 1)), ('t', 1)]))]))])

        from mindsdb.api.mongo.responders import responders
        respondersCollection.responders += responders


def run_server(config):
    SocketServer.TCPServer.allow_reuse_address = True
    with MongoServer(config) as srv:
        srv.serve_forever()



from typing import Optional

import dill
import pandas as pd

from autogluon.tabular import TabularPredictor
from type_infer.api import infer_types

from mindsdb.integrations.libs.base import BaseMLEngine
from mindsdb.utilities import log
from .config import ClassificationConfig, RegressionConfig


logger = log.getLogger(__name__)


class AutoGluonHandler(BaseMLEngine):
    name = "autogluon"

    def create(self, target: str, df: Optional[pd.DataFrame] = None, args: Optional[dict] = None) -> None:
        config_args = args['using']

        target_dtype = infer_types(df, 0).to_dict()["dtypes"][target]

        if target_dtype in ['binary', 'categorical', 'tags']:
            config = ClassificationConfig(**config_args)

            model = TabularPredictor(label=target, )
        elif target_dtype in ['integer', 'float', 'quantity']:
            config = RegressionConfig(**config_args)

            model = TabularPredictor(label=target)

        else:
            raise Exception('This task is not supported!')

        model.fit(df, **vars(config))
        self.model_storage.file_set('model', dill.dumps(model))
        self.model_storage.json_set('args', args)

    def predict(self, df: Optional[pd.DataFrame] = None, args: Optional[dict] = None) -> None:
        model = dill.loads(self.model_storage.file_get('model'))
        df = df.drop('__mindsdb_row_id', axis=1)

        predictions = model.predict(df)

        args = self.model_storage.json_get('args')
        df[args['target']] = predictions

        return df

    def describe(self, attribute: Optional[str] = None) -> pd.DataFrame:

        args = self.model_storage.json_get("args")

        if attribute == "args":
            return pd.DataFrame(args.items(), columns=["key", "value"])



import pandas as pd
from typing import List
from mindsdb.integrations.libs.api_handler import APITable
from mindsdb.integrations.utilities.handlers.query_utilities import SELECTQueryParser, SELECTQueryExecutor
from mindsdb.utilities import log
from mindsdb_sql.parser import ast

logger = log.getLogger(__name__)


class OilPriceLatestTable(APITable):
    """The Latest Oil Price Table implementation"""

    def select(self, query: ast.Select) -> pd.DataFrame:
        """Pulls data from the https://docs.oilpriceapi.com/guide/#prices-latest" API

        Parameters
        ----------
        query : ast.Select
           Given SQL SELECT query

        Returns
        -------
        pd.DataFrame
            latest oil price matching the query

        Raises
        ------
        ValueError
            If the query contains an unsupported condition
        """

        select_statement_parser = SELECTQueryParser(
            query,
            'latest_price',
            self.get_columns()
        )

        selected_columns, where_conditions, order_by_conditions, result_limit = select_statement_parser.parse_query()

        search_params = {}
        subset_where_conditions = []

        for op, arg1, arg2 in where_conditions:
            if arg1 == 'by_type':
                if op == '=':
                    search_params["by_type"] = arg2
                else:
                    raise NotImplementedError("Only '=' operator is supported for by_type column.")

                if not self.handler.client._is_valid_by_type(arg2):
                    raise ValueError("Unknown value for `by_type` parameter. The allowed values are - " + self.handler.client.valid_values_by_type)

            elif arg1 == 'by_code':
                if op == '=':
                    search_params["by_code"] = arg2
                else:
                    raise NotImplementedError("Only '=' operator is supported for by_code column.")

                if not self.handler.client._is_valid_by_code(arg2):
                    raise ValueError("Unknown value for `by_code` parameter. The allowed values are - " + self.handler.client.valid_values_by_code)

            elif arg1 in self.get_columns():
                subset_where_conditions.append([op, arg1, arg2])

        latest_price_df = pd.DataFrame(columns=self.get_columns())

        response = self.handler.client.get_latest_price(search_params.get("by_type"), search_params.get("by_code"))

        self.check_res(res=response)

        content = response["content"]

        latest_price_df = pd.json_normalize(content["data"])

        select_statement_executor = SELECTQueryExecutor(
            latest_price_df,
            selected_columns,
            subset_where_conditions,
            order_by_conditions,
            result_limit
        )

        latest_price_df = select_statement_executor.execute_query()

        return latest_price_df

    def check_res(self, res):
        if res["code"] != 200:
            raise Exception("Error fetching results - " + res["error"])

    def get_columns(self) -> List[str]:
        """Gets all columns to be returned in pandas DataFrame responses

        Returns
        -------
        List[str]
            List of columns
        """

        return [
            "price",
            "formatted",
            "currency",
            "code",
            "created_at",
            "type"
        ]


class OilPricePastDayPriceTable(APITable):
    """The Past Day Oil Price Table implementation"""

    def select(self, query: ast.Select) -> pd.DataFrame:
        """Pulls data from the https://docs.oilpriceapi.com/guide/#prices-past-day" API

        Parameters
        ----------
        query : ast.Select
           Given SQL SELECT query

        Returns
        -------
        pd.DataFrame
            past day oil price matching the query

        Raises
        ------
        ValueError
            If the query contains an unsupported condition
        """

        select_statement_parser = SELECTQueryParser(
            query,
            'past_day_price',
            self.get_columns()
        )

        selected_columns, where_conditions, order_by_conditions, result_limit = select_statement_parser.parse_query()

        search_params = {}
        subset_where_conditions = []

        for op, arg1, arg2 in where_conditions:
            if arg1 == 'by_type':
                if op == '=':
                    search_params["by_type"] = arg2
                else:
                    raise NotImplementedError("Only '=' operator is supported for by_type column.")

                if not self.handler.client._is_valid_by_type(arg2):
                    raise ValueError("Unknown value for `by_type` parameter. The allowed values are - " + self.handler.client.valid_values_by_type)

            elif arg1 == 'by_code':
                if op == '=':
                    search_params["by_code"] = arg2
                else:
                    raise NotImplementedError("Only '=' operator is supported for by_code column.")

                if not self.handler.client._is_valid_by_code(arg2):
                    raise ValueError("Unknown value for `by_code` parameter. The allowed values are - " + self.handler.client.valid_values_by_code)

            elif arg1 in self.get_columns():
                subset_where_conditions.append([op, arg1, arg2])

        price_df = pd.DataFrame(columns=self.get_columns())

        response = self.handler.client.get_price_past_day(search_params.get("by_type"), search_params.get("by_code"))

        self.check_res(res=response)

        content = response["content"]

        price_df = pd.json_normalize(content["data"]["prices"])

        select_statement_executor = SELECTQueryExecutor(
            price_df,
            selected_columns,
            subset_where_conditions,
            order_by_conditions,
            result_limit
        )

        price_df = select_statement_executor.execute_query()

        return price_df

    def check_res(self, res):
        if res["code"] != 200:
            raise Exception("Error fetching results - " + res["error"])

    def get_columns(self) -> List[str]:
        """Gets all columns to be returned in pandas DataFrame responses

        Returns
        -------
        List[str]
            List of columns
        """

        return [
            "price",
            "formatted",
            "currency",
            "code",
            "created_at",
            "type"
        ]



"""
*******************************************************
 * Copyright (C) 2017 MindsDB Inc. <copyright@mindsdb.com>
 *
 * This file is part of MindsDB Server.
 *
 * MindsDB Server can not be copied and/or distributed without the express
 * permission of MindsDB Inc
 *******************************************************
"""
import re
import inspect
from textwrap import dedent

from mindsdb_sql import parse_sql
from mindsdb_sql.planner.steps import (
    ApplyTimeseriesPredictorStep,
    ApplyPredictorRowStep,
    ApplyPredictorStep,
)

from mindsdb_sql.exceptions import PlanningException
from mindsdb_sql.render.sqlalchemy_render import SqlalchemyRender
from mindsdb_sql.planner import query_planner

from mindsdb.api.executor.utilities.sql import query_df, get_query_models
from mindsdb.interfaces.model.functions import get_model_record
from mindsdb.api.executor.exceptions import (
    BadTableError,
    UnknownError,
    LogicError,
)
import mindsdb.utilities.profiler as profiler
from mindsdb.utilities.fs import create_process_mark, delete_process_mark

from . import steps
from .result_set import ResultSet, Column
from . steps.base import BaseStepCall

superset_subquery = re.compile(r'from[\s\n]*(\(.*\))[\s\n]*as[\s\n]*virtual_table', flags=re.IGNORECASE | re.MULTILINE | re.S)


class SQLQuery:

    step_handlers = {}

    def __init__(self, sql, session, execute=True, database=None):
        self.session = session

        if database is not None:
            self.database = database
        else:
            self.database = session.database

        self.context = {
            'database': None if self.database == '' else self.database.lower(),
            'row_id': 0
        }

        self.columns_list = None
        self.steps_data = []

        self.planner = None
        self.parameters = []
        self.fetched_data = None

        self.outer_query = None

        if isinstance(sql, str):
            # region workaround for subqueries in superset
            if 'as virtual_table' in sql.lower():
                subquery = re.findall(superset_subquery, sql)
                if isinstance(subquery, list) and len(subquery) == 1:
                    subquery = subquery[0]
                    self.outer_query = sql.replace(subquery, 'dataframe')
                    sql = subquery.strip('()')
            # endregion
            self.query = parse_sql(sql, dialect='mindsdb')
            self.context['query_str'] = sql
        else:
            self.query = sql
            renderer = SqlalchemyRender('mysql')
            try:
                self.context['query_str'] = renderer.get_string(self.query, with_failback=True)
            except Exception:
                self.context['query_str'] = str(self.query)

        self.create_planner()

        if execute:
            self.prepare_query(prepare=False)
            self.execute_query()

    @classmethod
    def register_steps(cls):

        cls.step_handlers = {}
        for _, cl in inspect.getmembers(steps):
            if inspect.isclass(cl) and issubclass(cl, BaseStepCall):
                if cl.bind is not None:
                    step_name = cl.bind.__name__
                    cls.step_handlers[step_name] = cl

    @profiler.profile()
    def create_planner(self):
        databases = self.session.database_controller.get_list()

        predictor_metadata = []

        query_tables = get_query_models(self.query, default_database=self.database)

        for project_name, table_name, table_version in query_tables:
            args = {
                'name': table_name,
                'project_name': project_name
            }
            if table_version is not None:
                args['active'] = None
                args['version'] = table_version

            model_record = get_model_record(**args)
            if model_record is None:
                # check if it is an agent
                try:
                    agent = self.session.agents_controller.get_agent(table_name, project_name)
                except ValueError:
                    continue
                if agent is not None:
                    model = self.session.model_controller.get_model(
                        agent.model_name,
                        project_name=project_name
                    )

                    predictor = {
                        'name': table_name,
                        'integration_name': project_name,  # integration_name,
                        'timeseries': False,
                        'id': model['id'],
                        'to_predict': model['predict'],
                    }
                    predictor_metadata.append(predictor)

                continue

            if model_record.status == 'error':
                dot_version_str = ''
                and_version_str = ''
                if table_version is not None:
                    dot_version_str = f'.{table_version}'
                    and_version_str = f' and version = {table_version}'

                raise BadTableError(dedent(f'''\
                    The model '{table_name}{dot_version_str}' cannot be used as it is currently in 'error' status.
                    For detailed information about the error, please execute the following command:

                        select error from information_schema.models where name = '{table_name}'{and_version_str};
                '''))

            ts_settings = model_record.learn_args.get('timeseries_settings', {})
            predictor = {
                'name': table_name,
                'integration_name': project_name,   # integration_name,
                'timeseries': False,
                'id': model_record.id,
                'to_predict': model_record.to_predict,
            }
            if ts_settings.get('is_timeseries') is True:
                window = ts_settings.get('window')
                order_by = ts_settings.get('order_by')
                if isinstance(order_by, list):
                    order_by = order_by[0]
                group_by = ts_settings.get('group_by')
                if isinstance(group_by, list) is False and group_by is not None:
                    group_by = [group_by]
                predictor.update({
                    'timeseries': True,
                    'window': window,
                    'horizon': ts_settings.get('horizon'),
                    'order_by_column': order_by,
                    'group_by_columns': group_by
                })

            predictor['model_types'] = model_record.data.get('dtypes', {})

            predictor_metadata.append(predictor)

        database = None if self.database == '' else self.database.lower()

        self.context['predictor_metadata'] = predictor_metadata
        self.planner = query_planner.QueryPlanner(
            self.query,
            integrations=databases,
            predictor_metadata=predictor_metadata,
            default_namespace=database,
        )

    def fetch(self, view='list'):
        data = self.fetched_data

        if view == 'dataframe':
            result = data.to_df()
        else:
            result = data.get_records_raw()

        return {
            'success': True,
            'result': result
        }

    def prepare_query(self, prepare=True):
        if prepare:
            # it is prepared statement call
            try:
                for step in self.planner.prepare_steps(self.query):
                    data = self.execute_step(step)
                    step.set_result(data)
                    self.steps_data.append(data)
            except PlanningException as e:
                raise LogicError(e)

            statement_info = self.planner.get_statement_info()

            self.columns_list = []
            for col in statement_info['columns']:
                self.columns_list.append(
                    Column(
                        database=col['ds'],
                        table_name=col['table_name'],
                        table_alias=col['table_alias'],
                        name=col['name'],
                        alias=col['alias'],
                        type=col['type']
                    )
                )

            self.parameters = [
                Column(
                    name=col['name'],
                    alias=col['alias'],
                    type=col['type']
                )
                for col in statement_info['parameters']
            ]

    def execute_query(self, params=None):
        if self.fetched_data is not None:
            # no need to execute
            return

        process_mark = None
        try:
            steps = list(self.planner.execute_steps(params))
            steps_classes = (x.__class__ for x in steps)
            predict_steps = (ApplyPredictorRowStep, ApplyPredictorStep, ApplyTimeseriesPredictorStep)
            if any(s in predict_steps for s in steps_classes):
                process_mark = create_process_mark('predict')
            for step in steps:
                with profiler.Context(f'step: {step.__class__.__name__}'):
                    data = self.execute_step(step)
                step.set_result(data)
                self.steps_data.append(data)
        except PlanningException as e:
            raise LogicError(e)
        except Exception as e:
            raise e
        finally:
            if process_mark is not None:
                delete_process_mark('predict', process_mark)

        # save updated query
        self.query = self.planner.query

        # there was no executing
        if len(self.steps_data) == 0:
            return

        try:
            if self.outer_query is not None:
                # workaround for subqueries in superset. remove it?
                # +++
                # ???

                result = self.steps_data[-1]
                df = result.to_df()

                df2 = query_df(df, self.outer_query)

                result2 = ResultSet().from_df(df2, database='', table_name='')

                self.columns_list = result2.columns
                self.fetched_data = result2

            else:
                result = self.steps_data[-1]
                self.fetched_data = result
        except Exception as e:
            raise UnknownError("error in preparing result query step") from e

        try:
            if hasattr(self, 'columns_list') is False:
                # how it becomes False?
                self.columns_list = self.fetched_data.columns

            if self.columns_list is None:
                self.columns_list = self.fetched_data.columns

            for col in self.fetched_data.find_columns('__mindsdb_row_id'):
                self.fetched_data.del_column(col)

        except Exception as e:
            raise UnknownError("error in column list step") from e

    def execute_step(self, step):
        cls_name = step.__class__.__name__
        handler = self.step_handlers.get(cls_name)
        if handler is None:
            raise UnknownError(f"Unknown step: {cls_name}")

        return handler(self).call(step)


SQLQuery.register_steps()



import psycopg
from psycopg.postgres import types
from psycopg.pq import ExecStatus
from pandas import DataFrame

from mindsdb_sql import parse_sql
from mindsdb_sql.render.sqlalchemy_render import SqlalchemyRender
from mindsdb_sql.parser.ast.base import ASTNode

from mindsdb.integrations.libs.base import DatabaseHandler
from mindsdb.utilities import log
from mindsdb.integrations.libs.response import (
    HandlerStatusResponse as StatusResponse,
    HandlerResponse as Response,
    RESPONSE_TYPE
)
import mindsdb.utilities.profiler as profiler

logger = log.getLogger(__name__)


class PostgresHandler(DatabaseHandler):
    """
    This handler handles connection and execution of the PostgreSQL statements.
    """
    name = 'postgres'

    @profiler.profile('init_pg_handler')
    def __init__(self, name=None, **kwargs):
        super().__init__(name)
        self.parser = parse_sql
        self.connection_args = kwargs.get('connection_data')
        self.dialect = 'postgresql'
        self.database = self.connection_args.get('database')
        self.renderer = SqlalchemyRender('postgres')

        self.connection = None
        self.is_connected = False

    def __del__(self):
        if self.is_connected:
            self.disconnect()

    @profiler.profile()
    def connect(self):
        """
        Establishes a connection to a PostgreSQL database.

        Raises:
            psycopg.Error: If an error occurs while connecting to the PostgreSQL database.

        Returns:
            psycopg.Connection: A connection object to the PostgreSQL database.
        """
        if self.is_connected:
            return self.connection

        config = {
            'host': self.connection_args.get('host'),
            'port': self.connection_args.get('port'),
            'user': self.connection_args.get('user'),
            'password': self.connection_args.get('password'),
            'dbname': self.connection_args.get('database')
        }

        if self.connection_args.get('sslmode'):
            config['sslmode'] = self.connection_args.get('sslmode')

        # If schema is not provided set public as default one
        if self.connection_args.get('schema'):
            config['options'] = f'-c search_path={self.connection_args.get("schema")},public'

        try:
            self.connection = psycopg.connect(**config, connect_timeout=10)
            self.is_connected = True
            return self.connection
        except psycopg.Error as e:
            logger.error(f'Error connecting to PostgreSQL {self.database}, {e}!')
            self.is_connected = False
            raise

    def disconnect(self):
        """
        Closes the connection to the PostgreSQL database if it's currently open.
        """
        if not self.is_connected:
            return
        self.connection.close()
        self.is_connected = False

    def check_connection(self) -> StatusResponse:
        """
        Checks the status of the connection to the PostgreSQL database.

        Returns:
            StatusResponse: An object containing the success status and an error message if an error occurs.
        """
        response = StatusResponse(False)
        need_to_close = not self.is_connected

        try:
            connection = self.connect()
            with connection.cursor() as cur:
                # Execute a simple query to test the connection
                cur.execute('select 1;')
            response.success = True
        except psycopg.Error as e:
            logger.error(f'Error connecting to PostgreSQL {self.database}, {e}!')
            response.error_message = str(e)

        if response.success and need_to_close:
            self.disconnect()
        elif not response.success and self.is_connected:
            self.is_connected = False

        return response

    def _cast_dtypes(self, df: DataFrame, description: list) -> None:
        """
        Cast df dtypes basing on postgres types
            Note:
                Date types casting is not provided because of there is no issues (so far).
                By default pandas will cast postgres date types to:
                 - date -> object
                 - time -> object
                 - timetz -> object
                 - timestamp -> datetime64[ns]
                 - timestamptz -> datetime64[ns, {tz}]

            Args:
                df (DataFrame)
                description (list): psycopg cursor description
        """
        types_map = {
            'int2': 'int16',
            'int4': 'int32',
            'int8': 'int64',
            'numeric': 'float64',
            'float4': 'float32',
            'float8': 'float64'
        }
        for column_index, _ in enumerate(df.columns):
            col = df.iloc[:, column_index]  # column names could be duplicated
            if str(col.dtype) == 'object':
                pg_type = types.get(description[column_index].type_code)
                if pg_type is not None and pg_type.name in types_map:
                    col = col.fillna(0)
                    try:
                        df.iloc[:, column_index] = col.astype(types_map[pg_type.name])
                    except ValueError as e:
                        logger.error(f'Error casting column {col.name} to {types_map[pg_type.name]}: {e}')

    @profiler.profile()
    def native_query(self, query: str) -> Response:
        """
        Executes a SQL query on the PostgreSQL database and returns the result.

        Args:
            query (str): The SQL query to be executed.

        Returns:
            Response: A response object containing the result of the query or an error message.
        """
        need_to_close = not self.is_connected

        connection = self.connect()
        with connection.cursor() as cur:
            try:
                cur.execute(query)
                if ExecStatus(cur.pgresult.status) == ExecStatus.COMMAND_OK:
                    response = Response(RESPONSE_TYPE.OK)
                else:
                    result = cur.fetchall()
                    df = DataFrame(
                        result,
                        columns=[x.name for x in cur.description]
                    )
                    self._cast_dtypes(df, cur.description)
                    response = Response(
                        RESPONSE_TYPE.TABLE,
                        df
                    )
                connection.commit()
            except Exception as e:
                logger.error(f'Error running query: {query} on {self.database}, {e}!')
                response = Response(
                    RESPONSE_TYPE.ERROR,
                    error_code=0,
                    error_message=str(e)
                )
                connection.rollback()

        if need_to_close:
            self.disconnect()
        return response

    @profiler.profile()
    def query(self, query: ASTNode) -> Response:
        """
        Executes a SQL query represented by an ASTNode and retrieves the data.

        Args:
            query (ASTNode): An ASTNode representing the SQL query to be executed.

        Returns:
            Response: The response from the `native_query` method, containing the result of the SQL query execution.
        """
        query_str = self.renderer.get_string(query, with_failback=True)
        logger.debug(f"Executing SQL query: {query_str}")
        return self.native_query(query_str)

    def get_tables(self) -> Response:
        """
        Retrieves a list of all non-system tables and views in the current schema of the PostgreSQL database.

        Returns:
            Response: A response object containing the list of tables and views, formatted as per the `Response` class.
        """
        query = """
            SELECT
                table_schema,
                table_name,
                table_type
            FROM
                information_schema.tables
            WHERE
                table_schema NOT IN ('information_schema', 'pg_catalog')
                and table_type in ('BASE TABLE', 'VIEW')
                and table_schema = current_schema()
        """
        return self.native_query(query)

    def get_columns(self, table_name: str) -> Response:
        """
        Retrieves column details for a specified table in the PostgreSQL database.

        Args:
            table_name (str): The name of the table for which to retrieve column information.

        Returns:
            Response: A response object containing the column details, formatted as per the `Response` class.
        Raises:
            ValueError: If the 'table_name' is not a valid string.
        """

        if not table_name or not isinstance(table_name, str):
            raise ValueError("Invalid table name provided.")

        query = f"""
            SELECT
                column_name as "Field",
                data_type as "Type"
            FROM
                information_schema.columns
            WHERE
                table_name = '{table_name}'
        """
        return self.native_query(query)



import json
from typing import Dict, List, Text, Any

import requests
from requests import Response


def create_an_entity(url: str, payload: Dict[Text, Any], bearer_token: str) -> None:
    """
    Makes a POST request to the given url
    Creates an entity for the given url, bearer token and payload

    url: URL to which the get request is made
    bearer_token: authentication token for the request
    payload: a dictionary which provides the metadata information regarding the entity being created

    Returns
    None
    """
    payload = json.dumps(payload, indent=2)
    headers = {
        "Authorization": f"Bearer {bearer_token}",
        "Content-Type": "application/json",
    }
    getresponse(
        request_type="POST", url=url, headers=headers, payload=payload, files=[]
    )


def get_an_entity(url: str, bearer_token: str) -> Any:
    """
    Makes a GET request to the given url
    Gets the entity for the given url and bearer token

    url: URL to which the get request is made
    bearer_token: authentication token for the request

    Returns
    Dictionary or list of dictionaries containing information/metadata corresponding to the entities
    """
    payload = {}
    headers = {
        "Authorization": f"Bearer {bearer_token}",
        "Content-Type": "application/json",
    }
    response = getresponse(
        request_type="GET", url=url, headers=headers, payload=payload, files=[]
    )
    response = response.json()["value"]
    return response


def update_an_entity(
    url: str, values_to_update: Dict[Text, Any], bearer_token: str
) -> None:
    """
    Makes a PATCH request to given url with the given values_to_update and bearer_token
    updates the entity with the provided values

    url: url provided by the user
    values_to_update: values that would be used to update the entity, would be passed in the payload
    Mostly would be a dictionary mapping fields to values
    bearer_token: authentication token passed in the header to make the request

    Returns
    None
    """
    payload = values_to_update
    payload = json.dumps(payload, indent=2)
    headers = {
        "Authorization": f"Bearer {bearer_token}",
        "Content-Type": "application/json",
    }
    getresponse(
        request_type="PATCH", url=url, headers=headers, payload=payload, files=[]
    )


def delete_an_entity(url: str, bearer_token: str):
    """
    Makes a DELETE request to the given url

    url: url string provided to which the request would be made
    bearer_token: authorization token which will be used to execute the request

    Returns
    None
    """
    payload = {}
    headers = {
        "Authorization": f"Bearer {bearer_token}",
        "Content-Type": "application/json",
    }
    getresponse(
        request_type="DELETE", url=url, headers=headers, payload=payload, files=[]
    )


def bearer_token_request(tenant_id: str, client_id: str, client_secret: str) -> Any:
    """
    Sends a request to login.microsoftonline.com for the given tenant to generate a bearer token
    which is then used in making graph api call

    tenant_id: tenant ID is a globally unique identifier (GUID) for your organization
    that is different from your organization or domain name
    client_id: client ID is a globally unique identifier (GUID) for your app registered in Entra
    client_secret: client secret is the password of the service principal or the app.

    Returns
    response: Dictionary containing bearer token and other information
    """
    url = f"https://login.microsoftonline.com/{tenant_id}/oauth2/token"

    payload = {
        "client_id": client_id,
        "client_secret": client_secret,
        "redirect_uri": "http://localhost",
        "grant_type": "client_credentials",
        "resource": "https://graph.microsoft.com",
    }
    files = []
    headers = {}

    response = getresponse(
        request_type="POST", url=url, headers=headers, payload=payload, files=files
    )
    response = response.json()
    return response


def getresponse(
    url: str,
    payload: Dict[Text, Any],
    files: List[Any],
    headers: Dict[Text, Any],
    request_type: str,
) -> Response:
    """
    Makes a standard HTTP request based on the params provided and returns the response.
    May raise an error if the response code does not indicate success

    url: url string provided to which the request would be made
    payload: the payload which supply additional info regarding the request
    files: the files that are needed to be passed in the request
    headers: additional information regarding the request
    may also indicate the type of content passed in the payload
    request_type: the request performs different actions based on the request type
                  DELETE/POST/PATCH/GET

    Returns
    response: may return based on the response code
    """
    response = requests.request(
        request_type, url, headers=headers, data=payload, files=files
    )
    status_code = response.status_code

    if 400 <= status_code <= 499:
        raise Exception("Client error: " + response.text)

    if 500 <= status_code <= 599:
        raise Exception("Server error: " + response.text)
    return response



from typing import Dict, Optional

import google.generativeai as genai
import pandas as pd
from mindsdb.integrations.libs.base import BaseMLEngine
from mindsdb.utilities import log

from mindsdb.integrations.utilities.handler_utils import get_api_key

logger = log.getLogger(__name__)


class GoogleGeminiHandler(BaseMLEngine):
    """
    Integration with the Google generative AI Python Library
    """

    name = "google_gemini"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.default_chat_model = "gemini-pro"
        self.supported_chat_models = ["gemini-pro"]
        self.generative = True
        self.connection = None

    def create(
        self,
        target: str,
        df: Optional[pd.DataFrame] = None,
        args: Optional[Dict] = None,
    ) -> None:
        if "model" not in args["using"]:
            args["using"]["model"] = self.default_chat_model
        elif args["using"]["model"] not in self.supported_chat_models:
            raise Exception(
                f"Invalid chat model. Please use one of {self.supported_chat_models}"
            )

        api_key = get_api_key('google_gemini', args["using"], self.engine_storage, strict=False)

        try:
            genai.configure(api_key=api_key)
            model = genai.GenerativeModel(args["using"]["model"])
            model.generate_content("test")
        except Exception as e:
            raise Exception(
                f"{e}: Invalid api key please check your api key"
            )

        args["using"]["google_gemini_api_key"] = api_key

        self.model_storage.json_set("args", args)

    def predict(
        self, df: Optional[pd.DataFrame] = None, args: Optional[Dict] = None
    ) -> pd.DataFrame:
        args = self.model_storage.json_get("args")
        api_key = args["using"]["google_gemini_api_key"]
        genai.configure(api_key=api_key)

        input_column = args["using"]["column"]
        if input_column not in df.columns:
            raise RuntimeError(f'Column "{input_column}" not found in input data')

        self.connection = genai.GenerativeModel(args["using"]["model"])
        result_df = pd.DataFrame()
        result_df["predictions"] = df[input_column].apply(self.predict_answer)
        result_df = result_df.rename(columns={"predictions": args["target"]})
        return result_df

    def predict_answer(self, text):
        """
        connects with google generative AI api to predict the answer for the particular question

        """

        completion = self.connection.generate_content(
            text
        )

        return completion.text



import time

import pandas as pd
from dataprep_ml.insights import analyze_dataset
from flask import request
from flask_restx import Resource
from mindsdb_sql import parse_sql
from mindsdb_sql.parser.ast import Constant
from pandas.core.frame import DataFrame

from mindsdb.api.http.namespaces.configs.analysis import ns_conf
from mindsdb.api.executor.utilities.sql import get_query_tables
from mindsdb.api.http.utils import http_error
from mindsdb.api.mysql.mysql_proxy.classes.fake_mysql_proxy import FakeMysqlProxy
from mindsdb.api.executor.data_types.response_type import (
    RESPONSE_TYPE as SQL_RESPONSE_TYPE,
)
from mindsdb.metrics.metrics import api_endpoint_metrics
from mindsdb.utilities import log

logger = log.getLogger(__name__)


def analyze_df(df: DataFrame) -> dict:
    if len(df) == 0:
        return {}

    cols = pd.Series(df.columns)

    # https://stackoverflow.com/questions/24685012/pandas-dataframe-renaming-multiple-identically-named-columns
    for dup in cols[cols.duplicated()].unique():
        cols[cols[cols == dup].index.values.tolist()] = [
            dup + "." + str(i) if i != 0 else dup for i in range(sum(cols == dup))
        ]

    # rename the columns with the cols list.
    df.columns = cols

    analysis = analyze_dataset(df)
    return analysis.to_dict()


@ns_conf.route("/query")
class QueryAnalysis(Resource):
    @ns_conf.doc("post_query_to_analyze")
    @api_endpoint_metrics('POST', '/analysis/query')
    def post(self):
        data = request.json
        query = data.get("query")
        context = data.get("context", {})
        limit = data.get("limit")
        if query is None or len(query) == 0:
            return http_error(400, "Missed query", "Need provide query to analyze")

        try:
            ast = parse_sql(query)
        except Exception as e:
            return http_error(500, "Wrong query", str(e))

        if limit is not None:
            ast.limit = Constant(limit)
            query = str(ast)

        mysql_proxy = FakeMysqlProxy()
        mysql_proxy.set_context(context)

        try:
            result = mysql_proxy.process_query(query)
        except Exception as e:
            import traceback

            logger.error(traceback.format_exc())
            return http_error(500, "Error", str(e))

        if result.type == SQL_RESPONSE_TYPE.ERROR:
            return http_error(500, f"Error {result.error_code}", result.error_message)
        if result.type != SQL_RESPONSE_TYPE.TABLE:
            return http_error(500, "Error", "Query does not return data")

        column_names = [x["name"] for x in result.columns]
        df = DataFrame(result.data, columns=column_names)
        analysis = analyze_df(df)

        query_tables = [
            table.to_string() for table in get_query_tables(ast)
        ]

        return {
            "analysis": analysis,
            "column_names": column_names,
            "row_count": len(result.data),
            "timestamp": time.time(),
            "tables": query_tables,
        }


@ns_conf.route("/data")
class DataAnalysis(Resource):
    @ns_conf.doc("post_data_to_analyze")
    @api_endpoint_metrics('POST', '/analysis/data')
    def post(self):
        payload = request.json
        column_names = payload.get("column_names")
        data = payload.get("data")

        timestamp = time.time()
        try:
            analysis = analyze_df(DataFrame(data, columns=column_names))
            return {"analysis": analysis, "timestamp": time.time()}
        except Exception as e:
            # Don't want analysis exceptions to show up on UI.
            # TODO: Fix analysis so it doesn't throw exceptions at all.
            return {
                'analysis': {},
                'timestamp': timestamp,
                'error': str(e)
            }



from mindsdb_sql import parse_sql
from mindsdb.integrations.libs.api_handler import APIHandler
from mindsdb.utilities import log
from mindsdb.integrations.handlers.google_analytics_handler.google_analytics_tables import ConversionEventsTable
from mindsdb.integrations.libs.response import (
    HandlerStatusResponse as StatusResponse,
    HandlerResponse as Response,
)

import json
import os

from google.analytics.admin_v1beta import AnalyticsAdminServiceClient
from google.oauth2 import service_account
from google.auth.transport.requests import Request
from googleapiclient.errors import HttpError

DEFAULT_SCOPES = ['https://www.googleapis.com/auth/analytics.readonly',
                  'https://www.googleapis.com/auth/analytics.edit',
                  'https://www.googleapis.com/auth/analytics'
                  ]

logger = log.getLogger(__name__)


class GoogleAnalyticsHandler(APIHandler):
    """A class for handling connections and interactions with the Google Analytics Admin API.

    Attributes:
        credentials_file (str): The path to the Google Auth Credentials file for authentication
        and interacting with the Google Analytics API on behalf of the user.

        Scopes (List[str], Optional): The scopes to use when authenticating with the Google Analytics API.
    """

    name = 'google_analytics'

    def __init__(self, name: str, **kwargs):
        super().__init__(name)
        self.page_size = 500
        self.connection_args = kwargs.get('connection_data', {})
        self.property_id = self.connection_args['property_id']
        if self.connection_args.get('credentials'):
            self.credentials_file = self.connection_args.pop('credentials')

        self.scopes = self.connection_args.get('scopes', DEFAULT_SCOPES)
        self.service = None
        self.is_connected = False
        conversion_events = ConversionEventsTable(self)
        self.conversion_events = conversion_events
        self._register_table('conversion_events', conversion_events)

    def _get_creds_json(self):
        if 'credentials_file' in self.connection_args:
            if os.path.isfile(self.connection_args['credentials_file']) is False:
                raise Exception("credentials_file must be a file path")
            with open(self.connection_args['credentials_file']) as source:
                info = json.load(source)
            return info
        elif 'credentials_json' in self.connection_args:
            info = self.connection_args['credentials_json']
            if not isinstance(info, dict):
                raise Exception("credentials_json has to be dict")
            info['private_key'] = info['private_key'].replace('\\n', '\n')
            return info
        else:
            raise Exception('Connection args have to content ether credentials_file or credentials_json')

    def create_connection(self):
        info = self._get_creds_json()
        creds = service_account.Credentials.from_service_account_info(info=info, scopes=self.scopes)

        if not creds or not creds.valid:
            if creds and creds.expired and creds.refresh_token:
                creds.refresh(Request())

        return AnalyticsAdminServiceClient(credentials=creds)

    def connect(self):
        """
        Authenticate with the Google Analytics Admin API using the credential file.

        Returns
        -------
        service: object
            The authenticated Google Analytics Admin API service object.
        """
        if self.is_connected is True:
            return self.service

        self.service = self.create_connection()
        self.is_connected = True

        return self.service

    def check_connection(self) -> StatusResponse:
        """
        Check connection to the handler.

        Returns
        -------
        response
            Status confirmation
        """
        response = StatusResponse(False)

        try:
            # Call the Google Analytics API
            service = self.connect()
            result = service.list_conversion_events(parent=f'properties/{self.property_id}')

            if result is not None:
                response.success = True
        except HttpError as error:
            response.error_message = f'Error connecting to Google Analytics api: {error}.'
            log.logger.error(response.error_message)

        if response.success is False and self.is_connected is True:
            self.is_connected = False

        return response

    def native_query(self, query_string: str = None) -> Response:
        ast = parse_sql(query_string, dialect="mindsdb")

        return self.query(ast)

    def get_api_url(self, endpoint):
        return f'{endpoint}/{self.property_id}'



from collections import OrderedDict

from mindsdb.integrations.libs.const import HANDLER_CONNECTION_ARG_TYPE as ARG_TYPE


connection_args = OrderedDict(
    user = {
        'type': ARG_TYPE.STR,
        'description': 'Rockset user name'
    },
    password = {
        'type': ARG_TYPE.PWD,
        'description': 'Rockset password',
        'secret': True
    },
    api_key = {
        'type': ARG_TYPE.STR,
        'description': 'Rockset API key'
    },
    api_server = {
        'type': ARG_TYPE.STR,
        'description': 'Rockset API server'
    },
    host = {
        'type': ARG_TYPE.STR,
        'description': 'Rockset host'
    },
    port = {
        'type': ARG_TYPE.INT,
        'description': 'Rockset port'
    },
    database = {
        'type': ARG_TYPE.STR,
        'description': 'Rockset database'
    }
)
connection_args_example = OrderedDict(
    user = 'rockset',
    password = 'rockset',
    api_key = "adkjf234rksjfa23waejf2",
    api_server = 'api-us-west-2.rockset.io',
    host = 'localhost',
    port = '3306',
    database = 'test'
)



from langchain_core.output_parsers import StrOutputParser
from langchain_core.prompts import ChatPromptTemplate
from langchain_core.runnables import RunnableParallel, RunnablePassthrough, RunnableSerializable


from mindsdb.integrations.utilities.rag.retrievers.auto_retriever import AutoRetriever
from mindsdb.integrations.utilities.rag.retrievers.multi_vector_retriever import MultiVectorRetriever


from mindsdb.integrations.utilities.rag.settings import RAGPipelineModel, DEFAULT_AUTO_META_PROMPT_TEMPLATE
from mindsdb.integrations.utilities.rag.vector_store import VectorStoreOperator


class LangChainRAGPipeline:
    """
    Builds a RAG pipeline using langchain LCEL components
    """

    def __init__(self, retriever_runnable, prompt_template, llm):

        self.retriever_runnable = retriever_runnable
        self.prompt_template = prompt_template
        self.llm = llm

    def with_returned_sources(self) -> RunnableSerializable:
        """
        Builds a RAG pipeline with returned sources
        :return:
        """

        def format_docs(docs):
            if isinstance(docs, str):
                # this is to handle the case where the retriever returns a string
                # instead of a list of documents e.g. SQLRetriever
                return docs
            return "\n\n".join(doc.page_content for doc in docs)

        prompt = ChatPromptTemplate.from_template(self.prompt_template)

        rag_chain_from_docs = (
                RunnablePassthrough.assign(context=(lambda x: format_docs(x["context"])))  # noqa: E126, E122
                | prompt
                | self.llm
                | StrOutputParser()
        )

        rag_chain_with_source = RunnableParallel(
            {"context": self.retriever_runnable, "question": RunnablePassthrough()}
        ).assign(answer=rag_chain_from_docs)

        return rag_chain_with_source

    @classmethod
    def from_retriever(cls, config: RAGPipelineModel):
        """
        Builds a RAG pipeline with returned sources using a simple vector store retriever
        :param config: RAGPipelineModel

        :return:
        """
        vector_store_operator = VectorStoreOperator(
            vector_store=config.vector_store,
            documents=config.documents,
            embeddings_model=config.embeddings_model,
            vector_store_config=config.vector_store_config
        )

        return cls(vector_store_operator.vector_store.as_retriever(), config.rag_prompt_template, config.llm)

    @classmethod
    def from_auto_retriever(cls, config: RAGPipelineModel):
        """
        Builds a RAG pipeline with returned sources using a AutoRetriever

        :param config: RAGPipelineModel

        :return:
        """

        if not config.retriever_prompt_template:
            config.retriever_prompt_template = DEFAULT_AUTO_META_PROMPT_TEMPLATE

        retriever_runnable = AutoRetriever(config=config).as_runnable()
        return cls(retriever_runnable, config.rag_prompt_template, config.llm)

    @classmethod
    def from_multi_vector_retriever(cls, config: RAGPipelineModel):
        """
        Builds a RAG pipeline with returned sources using a MultiVectorRetriever

        :param config: RAGPipelineModel
        """

        retriever_runnable = MultiVectorRetriever(config=config).as_runnable()
        return cls(retriever_runnable, config.rag_prompt_template, config.llm)



from typing import Dict, Optional

import pandas as pd

from mindsdb.integrations.handlers.writer_handler.evaluate import WriterEvaluator
from mindsdb.integrations.handlers.writer_handler.ingest import WriterIngestor
from mindsdb.integrations.handlers.writer_handler.rag import QuestionAnswerer
from mindsdb.integrations.handlers.writer_handler.settings import (
    DEFAULT_EMBEDDINGS_MODEL,
    EVAL_COLUMN_NAMES,
    WriterHandlerParameters,
    WriterLLMParameters,
)
from mindsdb.integrations.libs.base import BaseMLEngine
from mindsdb.integrations.utilities.datasets.dataset import (
    load_dataset,
    validate_dataframe,
)
from mindsdb.utilities import log

# these require no additional arguments

logger = log.getLogger(__name__)

def extract_llm_params(args):
    """extract llm params from input query args"""

    llm_params = {}
    for param in WriterLLMParameters.model_fields:
        if param in args:
            llm_params[param] = args.pop(param)

    args["llm_params"] = llm_params

    return args


class WriterHandler(BaseMLEngine):
    """
    WriterHandler is a MindsDB integration with Writer API LLMs that allows users to run question answering
    on their data by providing a question.

    The User is able to provide data that provides context for the questions, see create() method for more details.

    """

    name = "writer"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.generative = True

    @staticmethod
    def create_validation(target, args=None, **kwargs):
        if "using" not in args:
            raise Exception(
                "Writer engine requires a USING clause! Refer to its documentation for more details."
            )

    def create(
        self,
        target: str,
        df: pd.DataFrame = pd.DataFrame(),
        args: Optional[Dict] = None,
    ):
        """
        Dispatch is running embeddings and storing in a VectorDB, unless user already has embeddings persisted
        """

        input_args = extract_llm_params(args["using"])

        if "evaluate_dataset" not in input_args and df is not None:
            # if user doesn't provide an evaluation dataset, use the input df from create query
            input_args["evaluate_dataset"] = df.to_dict(orient="records")

        args = WriterHandlerParameters(**input_args)

        # create folder for vector store to persist embeddings
        args.vector_store_storage_path = self.engine_storage.folder_get(
            args.vector_store_folder_name
        )

        if df is not None and args.run_embeddings:
            if "context_columns" not in args:
                # if no context columns provided, use all columns in df
                logger.info("No context columns provided, using all columns in df")
                args.context_columns = df.columns.tolist()

            if "embeddings_model_name" not in args:
                logger.info(
                    f"No embeddings model provided in query, using default model: {DEFAULT_EMBEDDINGS_MODEL}"
                )

            ingestor = WriterIngestor(args=args, df=df)
            ingestor.embeddings_to_vectordb()

        else:
            logger.info("Skipping embeddings and ingestion into Chroma VectorDB")

        export_args = args.dict(exclude={"llm_params"})
        # 'callbacks' aren't json serializable, we do this to avoid errors
        export_args["llm_params"] = args.llm_params.dict(exclude={"callbacks"})

        # for mindsdb cloud, store data in shared file system
        # for cloud version of mindsdb to make it be usable by all mindsdb nodes
        self.engine_storage.folder_sync(args.vector_store_folder_name)

        self.model_storage.json_set("args", export_args)

    def predict(self, df: pd.DataFrame = None, args: dict = None):
        """
        Dispatch is performed depending on the underlying model type. Currently, only question answering
        is supported.
        """

        input_args = self.model_storage.json_get("args")
        args = WriterHandlerParameters(**input_args)

        if args.evaluation_type:
            # if user adds a WHERE clause with 'run_evaluation = true', run evaluation
            if "run_evaluation" in df.columns and df["run_evaluation"].tolist()[0]:
                return self.evaluate(args)
            else:
                logger.info(
                    "Skipping evaluation, running prediction only. "
                    "to run evaluation, add a WHERE clause with 'run_evaluation = true'"
                )

        args.vector_store_storage_path = self.engine_storage.folder_get(
            args.vector_store_folder_name
        )

        # get question answering results
        question_answerer = QuestionAnswerer(args=args)

        # get question from sql query
        # e.g. where question = 'What is the capital of France?'
        response = question_answerer.query(df["question"].tolist()[0])

        return pd.DataFrame(response)

    def evaluate(self, args: WriterHandlerParameters):

        if isinstance(args.evaluate_dataset, list):
            # if user provides a list of dicts, convert to dataframe and validate
            evaluate_df = validate_dataframe(
                pd.DataFrame(args.evaluate_dataset), EVAL_COLUMN_NAMES
            )
        else:
            evaluate_df = load_dataset(
                ml_task_type="question_answering", dataset_name=args.evaluate_dataset
            )
            args.context_columns = "context"

        if args.n_rows_evaluation:
            # if user specifies n_rows_evaluation in create, only use that many rows
            evaluate_df = evaluate_df.head(args.n_rows_evaluation)

        ingestor = WriterIngestor(df=evaluate_df, args=args)
        ingestor.embeddings_to_vectordb()

        evaluator = WriterEvaluator(args=args, df=evaluate_df, rag=QuestionAnswerer)
        df = evaluator.evaluate()

        evaluation_metrics = dict(
            mean_evaluation_metrics=evaluator.mean_evaluation_metrics,
            evaluation_df=df.to_dict(orient="records"),
        )

        self.model_storage.json_set("evaluation", evaluation_metrics)

        return df

    def describe(self, attribute: Optional[str] = None) -> pd.DataFrame:
        """
        Describe the model, or a specific attribute of the model
        """

        if attribute == "evaluation_output":
            evaluation = self.model_storage.json_get("evaluation")
            return pd.DataFrame(evaluation["evaluation_df"])
        elif attribute == "mean_evaluation_metrics":
            evaluation = self.model_storage.json_get("evaluation")
            return pd.DataFrame(evaluation["mean_evaluation_metrics"])
        else:
            raise ValueError(
                f"Attribute {attribute} not supported, try 'evaluation_output' or 'mean_evaluation_metrics'"
            )



import datetime
from pydantic import BaseModel


class EmailSearchOptions(BaseModel):
    """
    Represents IMAP search options to use when searching emails
    """
    # IMAP mailbox to search.
    mailbox: str = "INBOX"
    # Search by email subject.
    subject: str = None
    # Search based on who the email was sent to.
    to_field: str = None
    # Search based on who the email was from.
    from_field: str = None
    # Search based on when the email was received.
    since_date: datetime.date = None
    until_date: datetime.date = None
    # Search for all emails after this ID.
    since_email_id: str = None

    class Config:
        json_schema_extra = {
            "example": {
                "mailbox": "INBOX",
                "subject": "Test",
                "to_email": "example@example.com",
                "from_email": "hello@example.com",
                "since_date": "2021-01-01",
                "until_date": "2021-01-31",
                "since_email_id": "123"
            }

        }
        extra = "forbid"


class EmailConnectionDetails(BaseModel):
    """
    Represents the connection details for an email client
    """
    email: str
    password: str
    imap_server: str = "imap.gmail.com"
    smtp_server: str = "smtp.gmail.com"
    smtp_port: int = 587

    class Config:
        json_schema_extra = {
            "example": {
                "email": "joe@bloggs.com",
                "password": "password",
                "imap_server": "imap.gmail.com",
                "smtp_server": "smtp.gmail.com",
                "smtp_port": 587
            }
        }
        extra = "forbid"




"""
*******************************************************
 * Copyright (C) 2017 MindsDB Inc. <copyright@mindsdb.com>
 *
 * This file is part of MindsDB Server.
 *
 * MindsDB Server can not be copied and/or distributed without the express
 * permission of MindsDB Inc
 *******************************************************
"""
from mindsdb.api.executor.datahub.datahub import init_datahub
from mindsdb.utilities.config import Config
from mindsdb.interfaces.agents.agents_controller import AgentsController
from mindsdb.interfaces.model.model_controller import ModelController
from mindsdb.interfaces.database.database import DatabaseController
from mindsdb.interfaces.database.integrations import integration_controller
from mindsdb.interfaces.skills.skills_controller import SkillsController
from mindsdb.utilities import log

logger = log.getLogger(__name__)


class SessionController:
    """
    This class manages the server session
    """

    def __init__(self, api_type='http') -> object:
        """
        Initialize the session
        """
        self.api_type = api_type
        self.username = None
        self.auth = False
        self.logging = logger
        self.database = None

        self.config = Config()

        self.model_controller = ModelController()
        self.integration_controller = integration_controller
        self.database_controller = DatabaseController()
        self.skills_controller = SkillsController()

        # to prevent circular imports
        from mindsdb.interfaces.knowledge_base.controller import KnowledgeBaseController
        self.kb_controller = KnowledgeBaseController(self)

        self.datahub = init_datahub(self)
        self.agents_controller = AgentsController(self.datahub)

        self.prepared_stmts = {}
        self.packet_sequence_number = 0
        self.profiling = False
        self.predictor_cache = True
        self.show_secrets = False

    def inc_packet_sequence_number(self):
        self.packet_sequence_number = (self.packet_sequence_number + 1) % 256

    def register_stmt(self, statement):
        i = 1
        while i in self.prepared_stmts and i < 100:
            i = i + 1
        if i == 100:
            raise Exception("Too many unclosed queries")

        self.prepared_stmts[i] = dict(type=None, statement=statement, fetched=0)
        return i

    def unregister_stmt(self, stmt_id):
        del self.prepared_stmts[stmt_id]

    def to_json(self):
        return {
            "username": self.username,
            "auth": self.auth,
            "database": self.database,
            "prepared_stmts": self.prepared_stmts,
            "packet_sequence_number": self.packet_sequence_number,
        }

    def from_json(self, updated):
        for key in updated:
            setattr(self, key, updated[key])



import os
import io
import shutil
import tarfile
import hashlib
from pathlib import Path
from abc import ABC, abstractmethod
from typing import Union, Optional
from dataclasses import dataclass
from datetime import datetime
import threading

if os.name == 'posix':
    import fcntl

import psutil
from checksumdir import dirhash
try:
    import boto3
    from botocore.exceptions import ClientError as S3ClientError
except Exception:
    # Only required for remote storage on s3
    S3ClientError = FileNotFoundError
    pass


from mindsdb.utilities.config import Config
from mindsdb.utilities.context import context as ctx
import mindsdb.utilities.profiler as profiler
from mindsdb.utilities import log
from mindsdb.utilities.fs import safe_extract

logger = log.getLogger(__name__)


@dataclass(frozen=True)
class RESOURCE_GROUP:
    PREDICTOR = 'predictor'
    INTEGRATION = 'integration'
    TAB = 'tab'


RESOURCE_GROUP = RESOURCE_GROUP()


DIR_LOCK_FILE_NAME = 'dir.lock'
DIR_LAST_MODIFIED_FILE_NAME = 'last_modified.txt'
SERVICE_FILES_NAMES = (DIR_LOCK_FILE_NAME, DIR_LAST_MODIFIED_FILE_NAME)


def copy(src, dst):
    if os.path.isdir(src):
        if os.path.exists(dst):
            if dirhash(src) == dirhash(dst):
                return
        shutil.rmtree(dst, ignore_errors=True)
        shutil.copytree(src, dst, dirs_exist_ok=True)
    else:
        if os.path.exists(dst):
            if hashlib.md5(open(src, 'rb').read()).hexdigest() == hashlib.md5(open(dst, 'rb').read()).hexdigest():
                return
        try:
            os.remove(dst)
        except Exception:
            pass
        shutil.copy2(src, dst)


class BaseFSStore(ABC):
    """Base class for file storage
    """

    def __init__(self):
        self.config = Config()
        self.storage = self.config['paths']['storage']

    @abstractmethod
    def get(self, local_name, base_dir):
        """Copy file/folder from storage to {base_dir}

        Args:
            local_name (str): name of resource (file/folder)
            base_dir (str): path to copy the resource
        """
        pass

    @abstractmethod
    def put(self, local_name, base_dir):
        """Copy file/folder from {base_dir} to storage

        Args:
            local_name (str): name of resource (file/folder)
            base_dir (str): path to folder with the resource
        """
        pass

    @abstractmethod
    def delete(self, remote_name):
        """Delete file/folder from storage

        Args:
            remote_name (str): name of resource
        """
        pass


def get_dir_size(path: str):
    total = 0
    with os.scandir(path) as it:
        for entry in it:
            if entry.is_file():
                total += entry.stat().st_size
            elif entry.is_dir():
                total += get_dir_size(entry.path)
    return total


class AbsentFSStore(BaseFSStore):
    """Storage class that does not store anything. It is just a dummy.
    """
    def get(self, *args, **kwargs):
        pass

    def put(self, *args, **kwargs):
        pass

    def delete(self, *args, **kwargs):
        pass


class LocalFSStore(BaseFSStore):
    """Storage that stores files locally
    """

    def __init__(self):
        super().__init__()

    def get(self, local_name, base_dir):
        remote_name = local_name
        src = os.path.join(self.storage, remote_name)
        dest = os.path.join(base_dir, local_name)
        if not os.path.exists(dest) or get_dir_size(src) != get_dir_size(dest):
            copy(src, dest)

    def put(self, local_name, base_dir, compression_level=9):
        remote_name = local_name
        copy(
            os.path.join(base_dir, local_name),
            os.path.join(self.storage, remote_name)
        )

    def delete(self, remote_name):
        path = Path(self.storage).joinpath(remote_name)
        try:
            if path.is_file():
                path.unlink()
            else:
                shutil.rmtree(path)
        except FileNotFoundError:
            pass


class FileLock:
    """ file lock to make safe concurrent access to directory
        works as context
    """

    @staticmethod
    def lock_folder_path(relative_path: Path) -> Path:
        """ Args:
                relative_path (Path): path to resource directory relative to storage root

            Returns:
                Path: abs path to folder with lock file
        """
        config = Config()
        root_storage_path = Path(config.paths['root'])
        return config.paths['locks'] / relative_path.relative_to(root_storage_path)

    def __init__(self, relative_path: Path, mode: str = 'w'):
        """ Args:
                relative_path (Path): path to resource directory relative to storage root
                mode (str): lock for read (r) or write (w)
        """
        if os.name != 'posix':
            return

        self._local_path = FileLock.lock_folder_path(relative_path)
        self._lock_file_name = DIR_LOCK_FILE_NAME
        self._lock_file_path = self._local_path / self._lock_file_name
        self._mode = fcntl.LOCK_EX if mode == 'w' else fcntl.LOCK_SH

        if self._lock_file_path.is_file() is False:
            self._local_path.mkdir(parents=True, exist_ok=True)
            try:
                self._lock_file_path.write_text('')
            except Exception:
                pass

    def __enter__(self):
        if os.name != 'posix':
            return

        try:
            # On at least some systems, LOCK_EX can only be used if the file
            # descriptor refers to a file opened for writing.
            self._lock_fd = os.open(self._lock_file_path, os.O_RDWR | os.O_CREAT)
            fcntl.lockf(self._lock_fd, self._mode | fcntl.LOCK_NB)
        except (ValueError, FileNotFoundError):
            # file probably was deleted between open and lock
            logger.error(f'Cant accure lock on {self._local_path}')
            raise FileNotFoundError
        except BlockingIOError:
            logger.error(f'Directory is locked by another process: {self._local_path}')
            fcntl.lockf(self._lock_fd, self._mode)

    def __exit__(self, exc_type, exc_value, traceback):
        if os.name != 'posix':
            return

        try:
            fcntl.lockf(self._lock_fd, fcntl.LOCK_UN)
            os.close(self._lock_fd)
        except Exception:
            pass


class S3FSStore(BaseFSStore):
    """Storage that stores files in amazon s3
    """

    dt_format = '%d.%m.%y %H:%M:%S.%f'

    def __init__(self):
        super().__init__()
        if 's3_credentials' in self.config['permanent_storage']:
            self.s3 = boto3.client('s3', **self.config['permanent_storage']['s3_credentials'])
        else:
            self.s3 = boto3.client('s3')
        self.bucket = self.config['permanent_storage']['bucket']
        self._thread_lock = threading.Lock()

    def _get_remote_last_modified(self, object_name: str) -> datetime:
        """ get time when object was created/modified

            Args:
                object_name (str): name if file in bucket

            Returns:
                datetime
        """
        last_modified = self.s3.get_object_attributes(
            Bucket=self.bucket,
            Key=object_name,
            ObjectAttributes=['Checksum']
        )['LastModified']
        last_modified = last_modified.replace(tzinfo=None)
        return last_modified

    @profiler.profile()
    def _get_local_last_modified(self, base_dir: str, local_name: str) -> datetime:
        """ get 'last_modified' that saved locally

            Args:
                base_dir (str): path to base folder
                local_name (str): folder name

            Returns:
                datetime | None
        """
        last_modified_file_path = Path(base_dir) / local_name / DIR_LAST_MODIFIED_FILE_NAME
        if last_modified_file_path.is_file() is False:
            return None
        try:
            last_modified_text = last_modified_file_path.read_text()
            last_modified_datetime = datetime.strptime(last_modified_text, self.dt_format)
        except Exception:
            return None
        return last_modified_datetime

    @profiler.profile()
    def _save_local_last_modified(self, base_dir: str, local_name: str, last_modified: datetime):
        """ Save 'last_modified' to local folder

            Args:
                base_dir (str): path to base folder
                local_name (str): folder name
                last_modified (datetime)
        """
        last_modified_file_path = Path(base_dir) / local_name / DIR_LAST_MODIFIED_FILE_NAME
        last_modified_text = last_modified.strftime(self.dt_format)
        last_modified_file_path.write_text(last_modified_text)

    @profiler.profile()
    def _download(self, base_dir: str, remote_ziped_name: str,
                  local_ziped_path: str, last_modified: datetime = None):
        """ download file to s3 and unarchive it

            Args:
                base_dir (str)
                remote_ziped_name (str)
                local_ziped_path (str)
                last_modified (datetime, optional)
        """
        os.makedirs(base_dir, exist_ok=True)

        remote_size = self.s3.get_object_attributes(
            Bucket=self.bucket,
            Key=remote_ziped_name,
            ObjectAttributes=['ObjectSize']
        )['ObjectSize']
        if (remote_size * 2) > psutil.virtual_memory().available:
            fh = io.BytesIO()
            self.s3.download_fileobj(self.bucket, remote_ziped_name, fh)
            with tarfile.open(fileobj=fh) as tar:
                safe_extract(tar, path=base_dir)
        else:
            self.s3.download_file(self.bucket, remote_ziped_name, local_ziped_path)
            shutil.unpack_archive(local_ziped_path, base_dir)
            os.remove(local_ziped_path)

        # os.system(f'chmod -R 777 {base_dir}')

        if last_modified is None:
            last_modified = self._get_remote_last_modified(remote_ziped_name)
        self._save_local_last_modified(
            base_dir,
            remote_ziped_name.replace('.tar.gz', ''),
            last_modified
        )

    @profiler.profile()
    def get(self, local_name, base_dir):
        remote_name = local_name
        remote_ziped_name = f'{remote_name}.tar.gz'
        local_ziped_name = f'{local_name}.tar.gz'
        local_ziped_path = os.path.join(base_dir, local_ziped_name)

        folder_path = Path(base_dir) / local_name
        with FileLock(folder_path, mode='r'):
            local_last_modified = self._get_local_last_modified(base_dir, local_name)
            remote_last_modified = self._get_remote_last_modified(remote_ziped_name)
            if (
                local_last_modified is not None
                and local_last_modified == remote_last_modified
            ):
                return

        with FileLock(folder_path, mode='w'):
            self._download(
                base_dir,
                remote_ziped_name,
                local_ziped_path,
                last_modified=remote_last_modified
            )

    @profiler.profile()
    def put(self, local_name, base_dir, compression_level=9):
        # NOTE: This `make_archive` function is implemente poorly and will create an empty archive file even if
        # the file/dir to be archived doesn't exist or for some other reason can't be archived
        remote_name = local_name
        remote_zipped_name = f'{remote_name}.tar.gz'

        dir_path = Path(base_dir) / remote_name
        dir_size = sum(f.stat().st_size for f in dir_path.glob('**/*') if f.is_file())
        if (dir_size * 2) < psutil.virtual_memory().available:
            old_cwd = os.getcwd()
            fh = io.BytesIO()
            with self._thread_lock:
                os.chdir(base_dir)
                with tarfile.open(fileobj=fh, mode='w:gz', compresslevel=compression_level) as tar:
                    for path in dir_path.iterdir():
                        if path.is_file() and path.name in SERVICE_FILES_NAMES:
                            continue
                        tar.add(path.relative_to(base_dir))
                os.chdir(old_cwd)
            fh.seek(0)

            self.s3.upload_fileobj(
                fh,
                self.bucket,
                remote_zipped_name
            )
        else:
            shutil.make_archive(
                os.path.join(base_dir, remote_name),
                'gztar',
                root_dir=base_dir,
                base_dir=local_name
            )

            self.s3.upload_file(
                os.path.join(base_dir, remote_zipped_name),
                self.bucket,
                remote_zipped_name
            )
            os.remove(os.path.join(base_dir, remote_zipped_name))

        last_modified = self._get_remote_last_modified(remote_zipped_name)
        self._save_local_last_modified(base_dir, local_name, last_modified)

    @profiler.profile()
    def delete(self, remote_name):
        self.s3.delete_object(Bucket=self.bucket, Key=remote_name)


def FsStore():
    storage_location = Config()['permanent_storage']['location']
    if storage_location == 'absent':
        return AbsentFSStore()
    if storage_location == 'local':
        return LocalFSStore()
    if storage_location == 's3':
        return S3FSStore()
    raise Exception(f"Location: '{storage_location}' not supported")


class FileStorage:
    def __init__(self, resource_group: str, resource_id: int,
                 root_dir: str = 'content', sync: bool = True):
        """
            Args:
                resource_group (str)
                resource_id (int)
                root_dir (str)
                sync (bool)
        """

        self.resource_group = resource_group
        self.resource_id = resource_id
        self.root_dir = root_dir
        self.sync = sync

        self.folder_name = f'{resource_group}_{ctx.company_id}_{resource_id}'

        config = Config()
        self.fs_store = FsStore()
        self.content_path = Path(config['paths'][root_dir])
        self.resource_group_path = self.content_path / resource_group
        self.folder_path = self.resource_group_path / self.folder_name
        if self.folder_path.exists() is False:
            self.folder_path.mkdir(parents=True, exist_ok=True)

    @profiler.profile()
    def push(self, compression_level: int = 9):
        with FileLock(self.folder_path, mode='r'):
            self._push_no_lock(compression_level=compression_level)

    @profiler.profile()
    def _push_no_lock(self, compression_level: int = 9):
        self.fs_store.put(
            str(self.folder_name),
            str(self.resource_group_path),
            compression_level=compression_level
        )

    @profiler.profile()
    def push_path(self, path, compression_level: int = 9):
        # TODO implement push per element
        self.push(compression_level=compression_level)

    @profiler.profile()
    def pull(self):
        try:
            self.fs_store.get(
                str(self.folder_name),
                str(self.resource_group_path)
            )
        except (FileNotFoundError, S3ClientError):
            pass

    @profiler.profile()
    def pull_path(self, path):
        # TODO implement pull per element
        self.pull()

    @profiler.profile()
    def file_set(self, name, content):
        if self.sync is True:
            self.pull()

        with FileLock(self.folder_path, mode='w'):

            dest_abs_path = self.folder_path / name

            with open(dest_abs_path, 'wb') as fd:
                fd.write(content)

            if self.sync is True:
                self._push_no_lock()

    @profiler.profile()
    def file_get(self, name):
        if self.sync is True:
            self.pull()
        dest_abs_path = self.folder_path / name
        with FileLock(self.folder_path, mode='r'):
            with open(dest_abs_path, 'rb') as fd:
                return fd.read()

    @profiler.profile()
    def add(self, path: Union[str, Path], dest_rel_path: Optional[Union[str, Path]] = None):
        """Copy file/folder to persist storage

        Examples:
            Copy file 'args.json' to '{storage}/args.json'
            >>> fs.add('/path/args.json')

            Copy file 'args.json' to '{storage}/folder/opts.json'
            >>> fs.add('/path/args.json', 'folder/opts.json')

            Copy folder 'folder' to '{storage}/folder'
            >>> fs.add('/path/folder')

            Copy folder 'folder' to '{storage}/path/folder'
            >>> fs.add('/path/folder', 'path/folder')

        Args:
            path (Union[str, Path]): path to the resource
            dest_rel_path (Optional[Union[str, Path]]): relative path in storage to file or folder
        """
        if self.sync is True:
            self.pull()
        with FileLock(self.folder_path, mode='w'):

            path = Path(path)
            if isinstance(dest_rel_path, str):
                dest_rel_path = Path(dest_rel_path)

            if dest_rel_path is None:
                dest_abs_path = self.folder_path / path.name
            else:
                dest_abs_path = self.folder_path / dest_rel_path

            copy(
                str(path),
                str(dest_abs_path)
            )

            if self.sync is True:
                self._push_no_lock()

    @profiler.profile()
    def get_path(self, relative_path: Union[str, Path]) -> Path:
        """ Return path to file or folder

        Examples:
            get path to 'opts.json':
            >>> fs.get_path('folder/opts.json')
            ... /path/{storage}/folder/opts.json

        Args:
            relative_path (Union[str, Path]): Path relative to the storage folder

        Returns:
            Path: path to requested file or folder
        """
        if self.sync is True:
            self.pull()

        with FileLock(self.folder_path, mode='r'):
            if isinstance(relative_path, str):
                relative_path = Path(relative_path)
            # relative_path = relative_path.resolve()

            if relative_path.is_absolute():
                raise TypeError('FSStorage.get_path() got absolute path as argument')

            ret_path = self.folder_path / relative_path
            if not ret_path.exists():
                # raise Exception('Path does not exists')
                os.makedirs(ret_path)

        return ret_path

    def delete(self, relative_path: Union[str, Path] = '.'):
        path = (self.folder_path / relative_path).resolve()
        if isinstance(relative_path, str):
            relative_path = Path(relative_path)

        if relative_path.is_absolute():
            raise TypeError('FSStorage.delete() got absolute path as argument')

        # complete removal
        if path == self.folder_path.resolve():
            with FileLock(self.folder_path, mode='w'):
                self.fs_store.delete(self.folder_name)
                # NOTE on some fs .rmtree is not working if any file is open
                shutil.rmtree(str(self.folder_path))

            # region del file lock
            lock_folder_path = FileLock.lock_folder_path(self.folder_path)
            try:
                shutil.rmtree(lock_folder_path)
            except FileNotFoundError:
                logger.warning('Tried to delete file not found: %s', lock_folder_path)
            except Exception as e:
                raise e
            # endregion
            return

        if self.sync is True:
            self.pull()

        with FileLock(self.folder_path, mode='w'):
            if path.exists() is False:
                raise Exception('Path does not exists')

            if path.is_file():
                path.unlink()
            else:
                path.rmdir()

            if self.sync is True:
                self._push_no_lock()


class FileStorageFactory:
    def __init__(self, resource_group: str,
                 root_dir: str = 'content', sync: bool = True):
        self.resource_group = resource_group
        self.root_dir = root_dir
        self.sync = sync

    def __call__(self, resource_id: int):
        return FileStorage(
            resource_group=self.resource_group,
            root_dir=self.root_dir,
            sync=self.sync,
            resource_id=resource_id
        )



import json

from google.oauth2 import service_account
from google.cloud.spanner_dbapi.connection import connect, Connection
from google.cloud.sqlalchemy_spanner import SpannerDialect

import pandas as pd
from mindsdb_sql import parse_sql
from mindsdb_sql.parser.ast.base import ASTNode
from mindsdb_sql.parser.ast import CreateTable, Function
from mindsdb_sql.render.sqlalchemy_render import SqlalchemyRender

from mindsdb.integrations.libs.base import DatabaseHandler
from mindsdb.integrations.libs.response import RESPONSE_TYPE
from mindsdb.integrations.libs.response import HandlerResponse as Response
from mindsdb.integrations.libs.response import (
    HandlerStatusResponse as StatusResponse,
)
from mindsdb.utilities import log

logger = log.getLogger(__name__)

class CloudSpannerHandler(DatabaseHandler):
    """This handler handles connection and execution of the Cloud Spanner statements."""

    name = 'cloud_spanner'

    def __init__(self, name: str, **kwargs):
        super().__init__(name)
        self.parser = parse_sql
        self.connection_data = kwargs.get('connection_data')
        self.dialect = self.connection_data.get('dialect', 'googlesql')

        if self.dialect == 'postgres':
            self.renderer = SqlalchemyRender('postgres')
        else:
            self.renderer = SqlalchemyRender(SpannerDialect)

        self.connection = None
        self.is_connected = False

    def __del__(self):
        if self.is_connected is True:
            self.disconnect()

    def connect(self) -> Connection:
        """Connect to a Cloud Spanner database.

        Returns:
            Connection: The database connection.
        """

        if self.is_connected is True:
            return self.connection

        args = {
            'database_id': self.connection_data.get('database_id'),
            'instance_id': self.connection_data.get('instance_id'),
            'project': self.connection_data.get('project'),
            'credentials': self.connection_data.get('credentials'),
        }

        args['credentials'] = service_account.Credentials.from_service_account_info(
            json.loads(args['credentials'])
        )
        self.connection = connect(**args)
        self.is_connected = True

        return self.connection

    def disconnect(self):
        """Close the database connection."""

        if self.is_connected is False:
            return

        self.connection.close()
        self.is_connected = False

    def check_connection(self) -> StatusResponse:
        """Check the connection to the Cloud Spanner database.

        Returns:
            StatusResponse: Connection success status and error message if an error occurs.
        """

        response = StatusResponse(False)

        try:
            self.connect()
            response.success = True
        except Exception as e:
            logger.error(
                f'Error connecting to Cloud Spanner {self.connection_data["database_id"]}, {e}!'
            )
            response.error_message = str(e)
        finally:
            if response.success is True and self.is_connected:
                self.disconnect()
            if response.success is False and self.is_connected:
                self.is_connected = False

        return response

    def native_query(self, query: str) -> Response:
        """Execute a SQL query.

        Args:
            query (str): The SQL query to execute.

        Returns:
            Response: The query result.
        """

        connection = self.connect()
        cursor = connection.cursor()

        try:
            cursor.execute(query)

            # The cursor description check indicates if there are any results.
            # This is required as spanner_dbapi will fail on a fetchall() call on an empty cursor.
            if cursor.description:
                result = cursor.fetchall()
                response = Response(
                    RESPONSE_TYPE.TABLE,
                    data_frame=pd.DataFrame(
                        result, columns=[x[0] for x in cursor.description]
                    ),
                )
            else:
                response = Response(RESPONSE_TYPE.OK)

            connection.commit()
        except Exception as e:
            logger.error(
                f'Error running query: {query} on {self.connection_data["database_id"]}!'
            )
            response = Response(RESPONSE_TYPE.ERROR, error_message=str(e))

        cursor.close()
        if self.is_connected:
            self.disconnect()

        return response

    def query(self, query: ASTNode) -> Response:
        """Render and execute a SQL query.

        Args:
            query (ASTNode): The SQL query.

        Returns:
            Response: The query result.
        """

        # check primary key for table:
        if isinstance(query, CreateTable) and query.columns is not None:
            id_col = None
            has_primary = False
            for col in query.columns:
                if col.name.lower() == 'id':
                    id_col = col
                if col.is_primary_key:
                    has_primary = True
            # if no other primary keys use id
            if not has_primary and id_col:
                id_col.is_primary_key = True
                id_col.default = Function('GENERATE_UUID', args=[])

        query_str = self.renderer.get_string(query, with_failback=True)

        return self.native_query(query_str)

    def get_tables(self) -> Response:
        """Get a list of all the tables in the database.

        Returns:
            Response: Names of the tables in the database.
        """

        query = '''
            SELECT
              t.table_name
            FROM
              information_schema.tables AS t
            WHERE
              t.table_schema = ''
        '''
        result = self.native_query(query)
        df = result.data_frame
        result.data_frame = df.rename(columns={df.columns[0]: 'table_name'})
        return result

    def get_columns(self, table_name: str) -> Response:
        """Get details about a table.

        Args:
            table_name (str): Name of the table to retrieve details of.

        Returns:
            Response: Details of the table.
        """

        query = f'''
            SELECT
              t.column_name,
              t.spanner_type,
              t.is_nullable
            FROM
              information_schema.columns AS t
            WHERE
              t.table_name = '{table_name}'
        '''
        return self.native_query(query)



from collections import OrderedDict

from mindsdb.integrations.libs.const import HANDLER_CONNECTION_ARG_TYPE as ARG_TYPE


connection_args = OrderedDict(
    host={
        'type': ARG_TYPE.STR,
        'description': 'The host name or IP address of the Empress Embedded server.'
    },
    port={
        'type': ARG_TYPE.INT,
        'description': 'Specify port to connect to Empress Embedded server'
    },
    user={
        'type': ARG_TYPE.STR,
        'description': 'The user name used to authenticate with the Empress Embedded server.'
    },
    password={
        'type': ARG_TYPE.PWD,
        'description': 'The password to authenticate the user with the Empress Embedded server.',
        'secret': True
    },
    server={
        'type': ARG_TYPE.STR,
        'description': 'The server name used to authenticate with the Empress Embedded server.'
    },
    database={
        'type': ARG_TYPE.STR,
        'description': 'Specify database name to connect Empress Embedded server'
    },

)

connection_args_example = OrderedDict(
    host='localhost',
    port=6322,
    user='admin',
    password='password',
    server='test',
    database='test_db'
)



from datetime import datetime, date, timedelta
from decimal import Decimal
import numpy as np
import pandas as pd
import json


class CustomJSONEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, np.ndarray):
            return obj.tolist()
        if pd.isnull(obj):
            return None
        if isinstance(obj, timedelta):
            return str(obj)
        if isinstance(obj, datetime):
            return obj.strftime("%Y-%m-%d %H:%M:%S.%f")
        if isinstance(obj, date):
            return obj.strftime("%Y-%m-%d")
        if isinstance(obj, np.bool_):
            return bool(obj)
        if isinstance(obj, np.int8) or isinstance(obj, np.int16) or isinstance(obj, np.int32) or isinstance(obj, np.int64):
            return int(obj)
        if isinstance(obj, np.float16) or isinstance(obj, np.float32) or isinstance(obj, np.float64) or isinstance(obj, Decimal):
            return float(obj)

        return str(obj)



from mindsdb.integrations.libs.const import HANDLER_TYPE
from mindsdb.utilities import log

logger = log.getLogger(__name__)

from .__about__ import __description__ as description
from .__about__ import __version__ as version

try:
    from .mindsdb_inference_handler import MindsDBInferenceHandler as Handler

    import_error = None
except Exception as e:
    Handler = None
    import_error = e

title = "MindsDB Inference"
name = "mindsdb_inference"
type = HANDLER_TYPE.ML
icon_path = 'icon.svg'
permanent = False

__all__ = ["Handler", "version", "name", "type", "title", "description", "import_error", "icon_path"]


import unittest
from mindsdb_sql import parse_sql

from mindsdb.api.executor.data_types.response_type import (
    RESPONSE_TYPE,
)
from mindsdb.integrations.handlers.faunadb_handler.faunadb_handler import (
    FaunaDBHandler,
)


class FaunadbHandlerTest(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.kwargs = {
            "connection_data": {
                "fauna_secret": "fnAFQFQPZNAAUYkCYkdvozJsm9tH2VbX55AULhsH",
                "fauna_endpoint": "https://db.fauna.com:443/",
            }
        }
        cls.handler = FaunaDBHandler("test_faunadb_handler", **cls.kwargs)

    def test_0_connect(self):
        self.handler.connect()

    def test_1_check_connection(self):
        self.handler.check_connection()
        self.assertTrue(self.handler.is_connected)

    def test_2_select(self):
        query = parse_sql("SELECT * FROM books;")
        res = self.handler.query(query)
        assert res.type is RESPONSE_TYPE.TABLE

    def test_3_describe_db(self):
        res = self.handler.get_tables()
        assert res.type is RESPONSE_TYPE.TABLE


if __name__ == "__main__":
    unittest.main()



try:
    import torch.multiprocessing as mp
except Exception:
    import multiprocessing as mp

ctx = mp.get_context('spawn')


class HandlerProcess(ctx.Process):
    daemon = True

    def __init__(self, fn, *args, **kwargs):
        super(HandlerProcess, self).__init__(args=args, kwargs=kwargs)
        self.fn = fn

    def run(self):
        self.fn(*self._args, **self._kwargs)



from http import HTTPStatus

import json

from flask import request
from flask_restx import Resource
from sqlalchemy.exc import NoResultFound


from mindsdb.api.http.namespaces.configs.projects import ns_conf
from mindsdb.api.http.utils import http_error
from mindsdb.api.executor.controllers.session_controller import SessionController
from mindsdb.metrics.metrics import api_endpoint_metrics
from mindsdb.interfaces.model.functions import PredictorRecordNotFound
from mindsdb.interfaces.storage.db import Predictor
from mindsdb_sql import parse_sql
from mindsdb_sql.parser.dialects.mindsdb import CreatePredictor


@ns_conf.route('/<project_name>/models')
class ModelsList(Resource):
    @ns_conf.doc('list_models')
    @api_endpoint_metrics('GET', '/models')
    def get(self, project_name):
        ''' List all models '''
        session = SessionController()

        try:
            session.database_controller.get_project(project_name)
        except NoResultFound:
            return http_error(
                HTTPStatus.NOT_FOUND,
                'Project not found',
                f'Project name {project_name} does not exist')

        return session.model_controller.get_models(with_versions=True, project_name=project_name)

    @ns_conf.doc('train_model')
    @api_endpoint_metrics('POST', '/models')
    def post(self, project_name):
        '''Creates a new model and trains it'''
        session = SessionController()

        if 'query' not in request.json:
            return http_error(
                HTTPStatus.BAD_REQUEST,
                'Query required',
                'Missing "query" SQL statement')
        query = request.json['query']

        project_datanode = session.datahub.get(project_name)
        if project_datanode is None:
            return http_error(
                HTTPStatus.NOT_FOUND,
                'Project not found',
                f'Project name {project_name} does not exist')

        try:
            create_statement = parse_sql(query, dialect='mindsdb')
        except Exception:
            return http_error(
                HTTPStatus.BAD_REQUEST,
                'Invalid query string',
                f'SQL CREATE statement is invalid: {query}')

        if type(create_statement) is not CreatePredictor:
            return http_error(
                HTTPStatus.BAD_REQUEST,
                'Invalid CREATE SQL statement',
                f'SQL statement is not a CREATE model statement: {query}')

        model_name = create_statement.name.parts[1].lower()
        try:
            session.model_controller.get_model(model_name, project_name=project_name)
            return http_error(
                HTTPStatus.CONFLICT,
                'Model already exists',
                f'Model with name {model_name} already exists')
        except PredictorRecordNotFound:
            pass

        ml_integration = 'lightwood'
        if create_statement.using is not None:
            # Convert using to lowercase
            create_statement.using = {k.lower(): v for k, v in create_statement.using.items()}
            ml_integration = create_statement.using.pop("engine", ml_integration)

        try:
            ml_handler = session.integration_controller.get_ml_handler(ml_integration)
        except Exception:
            return http_error(
                HTTPStatus.NOT_FOUND,
                'ML handler not found',
                f'Cannot find ML handler with name {ml_integration}')

        try:
            model_df = session.model_controller.create_model(create_statement, ml_handler)
            # Consistent format with GET /projects/<project_name>/models/<model_name>
            return {
                'name': model_df.at[0, 'NAME'],
                'accuracy': None,
                'active': model_df.at[0, 'ACTIVE'],
                'version': model_df.at[0, 'VERSION'],
                'status': model_df.at[0, 'STATUS'],
                'predict': model_df.at[0, 'PREDICT'],
                'mindsdb_version': model_df.at[0, 'MINDSDB_VERSION'],
                'error': model_df.at[0, 'ERROR'],
                'fetch_data_query': model_df.at[0, 'SELECT_DATA_QUERY'],
                'problem_definition': model_df.at[0, 'TRAINING_OPTIONS']
            }, HTTPStatus.CREATED
        except Exception as e:
            return http_error(
                HTTPStatus.INTERNAL_SERVER_ERROR,
                'Unable to train model',
                f'Something went wrong while creating and training model {model_name}: {e}')


@ns_conf.route('/<project_name>/models/<model_name>')
@ns_conf.param('project_name', 'Name of the project')
@ns_conf.param('model_name', 'Name of the model')
class ModelResource(Resource):
    @ns_conf.doc('get_model')
    @api_endpoint_metrics('GET', '/models/model')
    def get(self, project_name, model_name):
        '''Get a model by name and version'''
        session = SessionController()

        project_datanode = session.datahub.get(project_name)
        if project_datanode is None:
            return http_error(
                HTTPStatus.NOT_FOUND,
                'Project not found',
                f'Project name {project_name} does not exist')

        name_no_version, version = Predictor.get_name_and_version(model_name)
        try:
            return session.model_controller.get_model(name_no_version, version=version, project_name=project_name)
        except PredictorRecordNotFound:
            return http_error(
                HTTPStatus.NOT_FOUND,
                'Model not found',
                f'Model with name {model_name} not found')

    @ns_conf.doc('update_model')
    @api_endpoint_metrics('PUT', '/models/model')
    def put(self, project_name, model_name):
        """Update model"""

        session = SessionController()

        project_datanode = session.datahub.get(project_name)
        if project_datanode is None:
            return http_error(
                HTTPStatus.NOT_FOUND,
                'Project not found',
                f'Project name {project_name} does not exist')

        if 'problem_definition' not in request.json:
            return http_error(
                HTTPStatus.BAD_REQUEST,
                'problem_definition required',
                'Missing "problem_definition" field')

        problem_definition = request.json['problem_definition']

        model_name, version = Predictor.get_name_and_version(model_name)

        session.model_controller.update_model(
            session,
            project_name,
            model_name,
            version=version,
            problem_definition=problem_definition,
        )
        return session.model_controller.get_model(model_name, version=version, project_name=project_name)

    @ns_conf.doc('delete_model')
    @api_endpoint_metrics('DELETE', '/models/model')
    def delete(self, project_name, model_name):
        '''Deletes a model by name'''

        session = SessionController()

        project_datanode = session.datahub.get(project_name)
        if project_datanode is None:
            return http_error(
                HTTPStatus.NOT_FOUND,
                'Project not found',
                f'Project name {project_name} does not exist')

        name_no_version, version = Predictor.get_name_and_version(model_name)
        try:
            session.model_controller.get_model(name_no_version, version=version, project_name=project_name)
        except PredictorRecordNotFound:
            return http_error(
                HTTPStatus.NOT_FOUND,
                'Model not found',
                f'Model with name {model_name} not found')

        try:
            session.model_controller.delete_model(name_no_version, project_name, version=version)
        except Exception as e:
            return http_error(
                HTTPStatus.INTERNAL_SERVER_ERROR,
                'Error deleting model',
                f'Something went wrong while deleting {model_name}: {e}')

        return '', HTTPStatus.NO_CONTENT


@ns_conf.route('/<project_name>/models/<model_name>/predict')
@ns_conf.param('project_name', 'Name of the project')
@ns_conf.param('model_name', 'Name of the model')
class ModelPredict(Resource):
    @ns_conf.doc('post_model_predict')
    @api_endpoint_metrics('POST', '/models/model/predict')
    def post(self, project_name, model_name):
        '''Call prediction'''

        name_no_version, version = Predictor.get_name_and_version(model_name)

        session = SessionController()
        project_datanode = session.datahub.get(project_name)
        if project_datanode is None:
            return http_error(
                HTTPStatus.NOT_FOUND,
                'Project not found',
                f'Project with name {project_name} does not exist')

        try:
            session.model_controller.get_model(name_no_version, version=version, project_name=project_name)
        except PredictorRecordNotFound:
            return http_error(
                HTTPStatus.NOT_FOUND,
                'Model not found',
                f'Model with name {model_name} not found')

        data = request.json['data']
        if isinstance(data, str):
            # Support object or serialized object.
            data = json.loads(data)
        params = request.json.get('params')

        predictions = project_datanode.predict(
            model_name=name_no_version,
            data=data,
            version=version,
            params=params,
        )

        return predictions.to_dict('records')


@ns_conf.route('/<project_name>/models/<model_name>/describe')
@ns_conf.param('project_name', 'Name of the project')
@ns_conf.param('model_name', 'Name of the model')
class ModelDescribe(Resource):
    @ns_conf.doc('describe_model')
    @api_endpoint_metrics('GET', '/models/model/describe')
    def get(self, project_name, model_name):
        '''Describes a model'''
        session = SessionController()

        project_datanode = session.datahub.get(project_name)
        if project_datanode is None:
            return http_error(
                HTTPStatus.NOT_FOUND,
                'Project not found',
                f'Project name {project_name} does not exist')

        name_no_version, version = Predictor.get_name_and_version(model_name)

        try:
            session.model_controller.get_model(name_no_version, version=version, project_name=project_name)
        except PredictorRecordNotFound:
            return http_error(
                HTTPStatus.NOT_FOUND,
                'Model not found',
                f'Model with name {model_name} not found')

        attribute = None
        if 'attribute' in request.json:
            attribute = request.json['attribute']

        try:
            description_df = session.model_controller.describe_model(
                session,
                project_name,
                name_no_version,
                attribute,
                version=version)
            return description_df.to_dict('records')
        except Exception:
            return http_error(
                HTTPStatus.BAD_REQUEST,
                'ML handler unsupported',
                f'ML handler for {model_name} does not support model description')



import pandas as pd
from mindsdb.integrations.libs.api_handler import APITable
from mindsdb.integrations.utilities.handlers.query_utilities import SELECTQueryParser, SELECTQueryExecutor, INSERTQueryParser
from mindsdb_sql.parser import ast


class LumaEventsTable(APITable):
    """The Luma Get Event Table implementation"""

    def select(self, query: ast.Select) -> pd.DataFrame:
        """Pulls data from the https://docs.lu.ma/reference/get-event-1 and https://docs.lu.ma/reference/calendar-list-events" API

        Parameters
        ----------
        query : ast.Select
           Given SQL SELECT query

        Returns
        -------
        pd.DataFrame
            Luma events

        Raises
        ------
        ValueError
            If the query contains an unsupported condition
        """

        select_statement_parser = SELECTQueryParser(
            query,
            'events',
            self.get_columns()
        )

        selected_columns, where_conditions, order_by_conditions, result_limit = select_statement_parser.parse_query()

        search_params = {}
        subset_where_conditions = []
        filter_flag = False
        for op, arg1, arg2 in where_conditions:
            if arg1 == 'event_id':
                if op == '=':
                    search_params["event_id"] = arg2
                    filter_flag = True
                else:
                    raise NotImplementedError("Only '=' operator is supported for event_id column.")
            elif arg1 in self.get_columns():
                subset_where_conditions.append([op, arg1, arg2])

        df = pd.DataFrame(columns=self.get_columns())

        if filter_flag:
            response = self.handler.luma_client.get_event(search_params["event_id"])
            event = response["content"]["event"]
            df = pd.json_normalize(event)
        else:
            response = self.handler.luma_client.list_events()
            content = response["content"]["entries"]
            events_only = [event["event"] for event in content]
            df = pd.json_normalize(events_only)

        select_statement_executor = SELECTQueryExecutor(
            df,
            selected_columns,
            subset_where_conditions,
            order_by_conditions,
            result_limit
        )

        df = select_statement_executor.execute_query()

        return df

    def get_columns(self) -> list:
        return ["api_id",
                "cover_url",
                "name",
                "description",
                "series_api_id",
                "start_at",
                "duration_interval",
                "end_at",
                "geo_latitude",
                "geo_longitude",
                "url",
                "timezone",
                "event_type",
                "user_api_id",
                "visibility",
                "geo_address_json.city",
                "geo_address_json.type",
                "geo_address_json.region",
                "geo_address_json.address",
                "geo_address_json.country",
                "geo_address_json.latitude",
                "geo_address_json.place_id",
                "geo_address_json.longitude",
                "geo_address_json.city_state",
                "geo_address_json.description",
                "geo_address_json.full_address"]

    def _parse_event_insert_data(self, event):
        data = {}
        data["name"] = event["name"]
        data["start_at"] = event["start_at"]
        data["timezone"] = event["timezone"]

        if "end_at" in event:
            data["end_at"] = event["end_at"]

        if "require_rsvp_approval" in event:
            data["require_rsvp_approval"] = event["require_rsvp_approval"]

        if "geo_latitude" in event:
            data["geo_latitude"] = event["geo_latitude"]

        if "geo_longitude" in event:
            data["geo_longitude"] = event["geo_longitude"]

        if "meeting_url" in event:
            data["meeting_url"] = event["meeting_url"]

        data["geo_address_json"] = {}

        if "geo_address_json_type" in event:
            data["geo_address_json"]["type"] = event["geo_address_json_type"]

        if "geo_address_json_place_id" in event:
            data["geo_address_json"]["place_id"] = event["geo_address_json_place_id"]

        if "geo_address_json_description" in event:
            data["geo_address_json"]["description"] = event["geo_address_json_description"]

        return data

    def insert(self, query: ast.ASTNode) -> None:
        """Inserts data into the API endpoint.

        Parameters
        ----------
        query : ast.Insert
           Given SQL INSERT query

        Returns
        -------
        None

        Raises
        ------
        ValueError
            If the query contains an unsupported condition
        """
        insert_statement_parser = INSERTQueryParser(
            query,
            supported_columns=["name", "start_at", "timezone", "end_at", "require_rsvp_approval", "geo_address_json_type", "geo_address_json_place_id", "geo_address_json_description", "geo_latitude", "geo_longitude", "meeting_url"],
            mandatory_columns=["name", "start_at", "timezone"],
            all_mandatory=False
        )

        event_data = insert_statement_parser.parse_query()

        for event in event_data:
            parsed_event_data = self._parse_event_insert_data(event)
            self.handler.luma_client.create_event(parsed_event_data)



from typing import Optional

import pandas as pd
import ibm_db_dbi as love
from ibm_db_sa.ibm_db import DB2Dialect_ibm_db as DB2Dialect
from mindsdb_sql import parse_sql
from mindsdb_sql.parser.ast.base import ASTNode
from mindsdb_sql.render.sqlalchemy_render import SqlalchemyRender

from mindsdb.integrations.libs.base import DatabaseHandler
from mindsdb.utilities import log
from mindsdb.integrations.libs.response import (
    HandlerStatusResponse as StatusResponse,
    HandlerResponse as Response,
    RESPONSE_TYPE,
)


logger = log.getLogger(__name__)


class DB2Handler(DatabaseHandler):
    name = "DB2"

    def __init__(self, name: str, connection_data: Optional[dict], **kwargs):
        """Initialize the handler
        Args:
            name (str): name of particular handler instance
            connection_data (dict): parameters for connecting to the database
            **kwargs: arbitrary keyword arguments.
        """
        super().__init__(name)

        self.kwargs = kwargs
        self.parser = parse_sql
        self.driver = "{IBM DB2 ODBC DRIVER}"
        self.database = connection_data["database"]
        self.user = connection_data["user"]
        self.password = connection_data["password"]
        self.schemaName = connection_data["schema_name"]
        self.host = connection_data["host"]
        self.port = connection_data["port"]
        self.connString = (
            "DRIVER={0};"
            "DATABASE={1};"
            "HOST={2};"
            "PORT={3};"
            "PROTOCOL={4};"
            "UID={5};"
            "PWD={6};"
        ).format(
            self.driver,
            self.database,
            self.host,
            self.port,
            "TCPIP",
            self.user,
            self.password,
        )

        self.connection = None
        self.is_connected = False

    def connect(self):
        """Set up any connections required by the handler
        Should return output of check_connection() method after attempting
        connection. Should switch self.is_connected.
        Returns:
            Connection Object
        """
        if self.is_connected is True:
            return self.connection

        try:
            self.connection = love.pconnect(self.connString, "", "")

            self.is_connected = True
        except Exception as e:
            logger.error(f"Error while connecting to {self.database}, {e}")

        return self.connection

    def disconnect(self):
        """Close any existing connections
        Should switch self.is_connected.
        """
        if self.is_connected is False:
            return
        try:
            self.connection.close()
            self.is_connected = False
        except Exception as e:
            logger.error(f"Error while disconnecting to {self.database}, {e}")

        return

    def check_connection(self) -> StatusResponse:
        """Check connection to the handler
        Returns:
            HandlerStatusResponse
        """
        responseCode = StatusResponse(False)
        need_to_close = self.is_connected is False

        try:
            self.connect()
            responseCode.success = True
        except Exception as e:
            logger.error(f"Error connecting to database {self.database}, {e}!")
            responseCode.error_message = str(e)
        finally:
            if responseCode.success is True and need_to_close:
                self.disconnect()
            if responseCode.success is False and self.is_connected is True:
                self.is_connected = False

        return responseCode

    def native_query(self, query: str) -> StatusResponse:
        """Receive raw query and act upon it somehow.
        Args:
            query (Any): query in native format (str for sql databases,
                dict for mongo, etc)
        Returns:
            HandlerResponse
        """
        need_to_close = self.is_connected is False
        query = query.upper()
        conn = self.connect()
        with conn.cursor() as cur:
            try:
                cur.execute(query)

                if cur._result_set_produced:
                    result = cur.fetchall()
                    response = Response(
                        RESPONSE_TYPE.TABLE,
                        data_frame=pd.DataFrame(
                            result, columns=[x[0] for x in cur.description]
                        ),
                    )
                else:
                    response = Response(RESPONSE_TYPE.OK)
                self.connection.commit()
            except Exception as e:
                logger.error(f"Error running query: {query} on {self.database}!")
                response = Response(RESPONSE_TYPE.ERROR, error_message=str(e))
                self.connection.rollback()

        if need_to_close is True:
            self.disconnect()

        return response

    def query(self, query: ASTNode) -> StatusResponse:
        """Receive query as AST (abstract syntax tree) and act upon it somehow.
        Args:
            query (ASTNode): sql query represented as AST. May be any kind
                of query: SELECT, INTSERT, DELETE, etc
        Returns:
            HandlerResponse
        """

        renderer = SqlalchemyRender(DB2Dialect)
        query_str = renderer.get_string(query, with_failback=True)
        return self.native_query(query_str)

    def get_tables(self) -> StatusResponse:
        """Return list of entities
        Return list of entities that will be accesible as tables.
        Returns:
            HandlerResponse: shoud have same columns as information_schema.tables
                (https://dev.mysql.com/doc/refman/8.0/en/information-schema-tables-table.html)
                Column 'TABLE_NAME' is mandatory, other is optional.
        """
        self.connect()

        result = self.connection.tables(self.schemaName)
        try:
            if result:
                response = Response(
                    RESPONSE_TYPE.TABLE,
                    data_frame=pd.DataFrame(
                        [result[i]["TABLE_NAME"] for i in range(len(result))],
                        columns=["TABLE_NAME"],
                    ),
                )
            else:
                response = Response(RESPONSE_TYPE.OK)

        except Exception as e:
            logger.error(f"Error running while getting table {e} on ")
            response = Response(RESPONSE_TYPE.ERROR, error_message=str(e))

        return response

    def get_columns(self, table_name: str) -> StatusResponse:
        """Returns a list of entity columns
        Args:
            table_name (str): name of one of tables returned by self.get_tables()
        Returns:
            HandlerResponse: shoud have same columns as information_schema.columns
                (https://dev.mysql.com/doc/refman/8.0/en/information-schema-columns-table.html)
                Column 'COLUMN_NAME' is mandatory, other is optional. Hightly
                recomended to define also 'DATA_TYPE': it should be one of
                python data types (by default it str).
        """

        self.connect()

        result = self.connection.columns(table_name=table_name)
        try:
            if result:
                response = Response(
                    RESPONSE_TYPE.TABLE,
                    data_frame=pd.DataFrame(
                        [result[i]["COLUMN_NAME"] for i in range(len(result))],
                        columns=["COLUMN_NAME"],
                    ),
                )
            else:
                response = Response(RESPONSE_TYPE.OK)

        except Exception as e:
            logger.error(f"Error running while getting table {e} on ")
            response = Response(RESPONSE_TYPE.ERROR, error_message=str(e))

        return response



"""
*******************************************************
 * Copyright (C) 2017 MindsDB Inc. <copyright@mindsdb.com>
 *
 * This file is part of MindsDB Server.
 *
 * MindsDB Server can not be copied and/or distributed without the express
 * permission of MindsDB Inc
 *******************************************************
"""


from mindsdb.api.mysql.mysql_proxy.data_types.mysql_packet import Packet
from mindsdb.api.mysql.mysql_proxy.data_types.mysql_datum import Datum


class STMTPrepareHeaderPacket(Packet):
    '''
    Implementation based on description:
    https://dev.mysql.com/doc/internals/en/com-stmt-prepare-response.html#packet-COM_STMT_PREPARE_OK
    '''

    def setup(self):
        self.status = Datum('int<1>', 0)
        self.stmt_id = Datum('int<4>', self._kwargs.get('stmt_id', 1))
        self.num_columns = Datum('int<2>', self._kwargs.get('num_columns', 0))
        self.num_params = Datum('int<2>', self._kwargs.get('num_params', 0))
        self.filler = Datum('int<1>', 0)
        self.warning_count = Datum('int<2>', 0)

    @property
    def body(self):
        order = [
            'status',
            'stmt_id',
            'num_columns',
            'num_params',
            'filler',
            'warning_count'
        ]
        string = b''
        for key in order:
            string += getattr(self, key).toStringPacket()

        self.setBody(string)
        return self._body



from typing import List, Optional, Union, Dict
from pydantic import BaseModel, Extra


class CompletionParameters(BaseModel):
    model: str  # The model to be used for the API call.
    prompt_template: Optional[str] = None  # Template for the prompt to be used.

    # Optional OpenAI params: see https://platform.openai.com/docs/api-reference/chat/create
    functions: List = []  # List of functions for advanced model operations.
    function_call: str = ""  # String to specify a particular function call, if needed.

    timeout: Optional[Union[float, int]] = None  # Timeout for the API request.
    temperature: Optional[float] = None  # Controls randomness: higher value means more random responses.
    top_p: Optional[float] = None  # Nucleus sampling: higher value means more diverse responses.
    n: Optional[int] = None  # Number of completions to generate for each prompt.
    stream: Optional[bool] = None  # Whether to stream responses or not.
    stop: Optional[str] = None  # Sequence at which the model should stop generating further tokens.
    max_tokens: Optional[float] = None  # Maximum number of tokens to generate in each completion.
    presence_penalty: Optional[float] = None  # Penalty for new tokens based on their existing presence in the text.
    frequency_penalty: Optional[float] = None  # Penalty for new tokens based on their frequency in the text.
    logit_bias: Optional[Dict] = None  # Adjusts the likelihood of specified tokens appearing in the completion.
    user: Optional[str] = None  # Identifier for the end-user, for usage tracking.

    # openai v1.0+ new params
    response_format: Optional[Dict] = None  # Format of the response, if specific formatting is required.
    seed: Optional[int] = None  # Random seed for deterministic completions.
    tools: Optional[List[str]] = None  # List of additional tools or features to be used.
    tool_choice: Optional[str] = None  # Specific tool to be used, if applicable.
    deployment_id: Optional[str] = None  # Identifier for the deployment.

    # set api_base, api_version, api_key
    base_url: Optional[str] = None  # Base URL of the API.
    api_version: Optional[str] = None  # Version of the API to be used.
    api_key: str  # API key for authentication.
    llm_model_list: Optional[List] = None  # List of models, API bases, keys, etc., for dynamic selection.

    class Config:
        extra = Extra.forbid
        arbitrary_types_allowed = True
        use_enum_values = True



from langchain_community.llms import Writer

from mindsdb.integrations.handlers.rag_handler.rag import RAGQuestionAnswerer
from mindsdb.integrations.handlers.writer_handler.settings import (
    WriterHandlerParameters,
)


class QuestionAnswerer(RAGQuestionAnswerer):
    def __init__(self, args: WriterHandlerParameters):

        super().__init__(args)

        self.llm = Writer(**args.llm_params.model_dump())



import os
import pytest
import pymssql
import pandas as pd
from mindsdb.integrations.handlers.mssql_handler.mssql_handler import SqlServerHandler
from mindsdb.integrations.libs.response import (
    HandlerResponse as Response,
    RESPONSE_TYPE,
)

HANDLER_KWARGS = {
    "connection_data": {
        "host": os.environ.get("MDB_TEST_MSSQL_HOST", "localhost"),
        "port": os.environ.get("MDB_TEST_MSSQL_PORT", "1433"),
        "user": os.environ.get("MDB_TEST_MSSQL_USER", "sa"),
        "password": os.environ.get("MDB_TEST_MSSQL_PASSWORD", "admin5678@"),
        "database": os.environ.get("MDB_TEST_MSSQL_DATABASE", "mdb_db_handler_test"),
    }
}


@pytest.fixture(scope="class")
def sql_server_handler():
    seed_db()
    handler = SqlServerHandler("test_sqlserver_handler", **HANDLER_KWARGS)
    yield handler
    handler.disconnect()


def seed_db():
    """Seed the test DB with some data"""

    # Connect to 'master' while we create our test DB
    conn_info = HANDLER_KWARGS["connection_data"].copy()
    conn_info["database"] = "master"
    db = pymssql.connect(**conn_info, autocommit=True)
    cursor = db.cursor()

    with open("mindsdb/integrations/handlers/mssql_handler/tests/seed.sql", "r") as f:
        for line in f.readlines():
            cursor.execute(line)
    cursor.close()
    db.close()


def check_valid_response(res):
    if res.resp_type == RESPONSE_TYPE.TABLE:
        assert res.data_frame is not None, "expected to have some data, but got None"
    assert (
        res.error_code == 0
    ), f"expected to have zero error_code, but got {res.error_code}"
    assert (
        res.error_message is None
    ), f"expected to have None in error message, but got {res.error_message}"


def get_table_names(sql_server_handler):
    res = sql_server_handler.get_tables()
    tables = res.data_frame
    assert tables is not None, "expected to have some tables in the db, but got None"
    assert (
        "table_name" in tables
    ), f"expected to get 'table_name' column in the response:\n{tables}"
    return list(tables["table_name"])


@pytest.mark.usefixtures("sql_server_handler")
class TestMSSQLHandlerConnect:
    def test_connect(self, sql_server_handler):
        sql_server_handler.connect()
        assert sql_server_handler.is_connected, "the handler has failed to connect"

    def test_check_connection(self, sql_server_handler):
        res = sql_server_handler.check_connection()
        assert res.success, res.error_message


class TestMSSQLHandlerDisconnect:
    def test_disconnect(self, sql_server_handler):
        sql_server_handler.disconnect()
        assert sql_server_handler.is_connected is False, "failed to disconnect"

    def test_check_connection(self, sql_server_handler):
        res = sql_server_handler.check_connection()
        assert res.success, res.error_message


@pytest.mark.usefixtures("sql_server_handler")
class TestMSSQLHandlerTables:
    table_for_creation = "test_mdb"

    def test_get_tables(self, sql_server_handler):
        res = sql_server_handler.get_tables()
        tables = res.data_frame
        assert (
            tables is not None
        ), "expected to have some tables in the db, but got None"
        assert (
            "table_name" in tables
        ), f"expected to get 'table_name' in the response but got: {tables}"
        assert (
            "test" in tables["table_name"].values
        ), "expected to have 'test' in the response."

    def test_get_columns(self, sql_server_handler):
        response = sql_server_handler.get_columns("test")
        assert response.type == RESPONSE_TYPE.TABLE, "expected a TABLE"
        assert len(response.data_frame) > 0, "expected > O columns"

        expected_columns = {
            "Field": ["col_one", "col_two", "col_three", "col_four"],
            "Type": ["int", "int", "float", "varchar"],
        }
        expected_df = pd.DataFrame(expected_columns)
        assert response.data_frame.equals(
            expected_df
        ), "response does not contain the expected columns"

    def test_create_table(self, sql_server_handler):
        query = f"""
        IF NOT EXISTS (SELECT * FROM sys.tables WHERE name = '{self.table_for_creation}')
        BEGIN
            CREATE TABLE {self.table_for_creation} (test_col INT)
        END
        """
        res = sql_server_handler.native_query(query)
        check_valid_response(res)
        tables = get_table_names(sql_server_handler)
        assert (
            self.table_for_creation in tables
        ), f"expected to have {self.table_for_creation} in database, but got: {tables}"

    def test_drop_table(self, sql_server_handler):
        query = f"DROP TABLE IF EXISTS {self.table_for_creation}"
        res = sql_server_handler.native_query(query)
        check_valid_response(res)
        tables = get_table_names(sql_server_handler)
        assert self.table_for_creation not in tables


@pytest.mark.usefixtures("sql_server_handler")
class TestMSSQLHandlerQuery:
    def test_select_native_query(self, sql_server_handler):
        query = "SELECT * FROM test"
        response = sql_server_handler.native_query(query)
        assert type(response) is Response
        assert response.resp_type == RESPONSE_TYPE.TABLE

        expected_data = {
            "col_one": [1, 2, 3],
            "col_two": [-1, -2, -3],
            "col_three": [0.1, 0.2, 0.3],
            "col_four": ["A", "B", "C"],
        }
        expected_df = pd.DataFrame(expected_data)
        assert response.data_frame.equals(
            expected_df
        ), "response does not contain the expected data"

    def test_select_query(self, sql_server_handler):
        limit = 3
        query = "SELECT * FROM test"
        res = sql_server_handler.query(query)
        check_valid_response(res)
        got_rows = res.data_frame.shape[0]
        want_rows = limit
        assert (
            got_rows == want_rows
        ), f"expected to have {want_rows} rows in response but got: {got_rows}"



from enum import Enum
from typing import List, Union

from langchain_community.vectorstores.chroma import Chroma
from langchain_community.vectorstores.pgvector import PGVector
from langchain_core.documents import Document
from langchain_core.embeddings import Embeddings
from langchain_core.language_models import BaseChatModel
from langchain_core.vectorstores import VectorStore
from langchain_core.stores import BaseStore
from langchain.text_splitter import TextSplitter
from pydantic import BaseModel

DEFAULT_COLLECTION_NAME = 'default_collection'

# Multi retriever specific
DEFAULT_ID_KEY = "doc_id"
DEFAULT_MAX_CONCURRENCY = 5

DEFAULT_CARDINALITY_THRESHOLD = 40
DEFAULT_CHUNK_SIZE = 1000
DEFAULT_CHUNK_OVERLAP = 50
DEFAULT_POOL_RECYCLE = 3600
DEFAULT_LLM_MODEL = "gpt-3.5-turbo"
DEFAULT_CONTENT_COLUMN_NAME = "body"
DEFAULT_DATASET_DESCRIPTION = "email inbox"
DEFAULT_TEST_TABLE_NAME = "test_email"
DEFAULT_VECTOR_STORE = Chroma
DEFAULT_AUTO_META_PROMPT_TEMPLATE = """
Below is a json representation of a table with information about {description}.
Return a JSON list with an entry for each column. Each entry should have
{{"name": "column name", "description": "column description", "type": "column data type"}}
\n\n{dataframe}\n\nJSON:\n
"""
DEFAULT_RAG_PROMPT_TEMPLATE = '''You are an assistant for
question-answering tasks. Use the following pieces of retrieved context
to answer the question. If you don't know the answer, just say that you
don't know. Use two sentences maximum and keep the answer concise.
Question: {question}
Context: {context}
Answer:'''

DEFAULT_QA_GENERATION_PROMPT_TEMPLATE = '''You are an assistant for
generating sample questions and answers from the given document and metadata. Given
a document and its metadata as context, generate a question and answer from that document and its metadata.

The document will be a string. The metadata will be a JSON string. You need
to parse the JSON to understand it.

Generate a question that requires BOTH the document and metadata to answer, if possible.
Otherwise, generate a question that requires ONLY the document to answer.

Return a JSON dictionary with the question and answer like this:
{{ "question": <the full generated question>, "answer": <the full generated answer> }}

Make sure the JSON string is valid before returning it. You must return the question and answer
in the specified JSON format no matter what.

Document: {document}
Metadata: {metadata}
Answer:'''


class MultiVectorRetrieverMode(Enum):
    """
    Enum for MultiVectorRetriever types.
    """
    SPLIT = "split"
    SUMMARIZE = "summarize"
    BOTH = "both"


class VectorStoreType(Enum):
    CHROMA = 'chromadb'
    PGVECTOR = 'pgvector'


vector_store_map = {
    VectorStoreType.CHROMA: Chroma,
    VectorStoreType.PGVECTOR: PGVector
}


class RetrieverType(Enum):
    VECTOR_STORE = 'vector_store'
    AUTO = 'auto'
    MULTI = 'multi'


class VectorStoreConfig(BaseModel):
    vector_store_type: VectorStoreType = VectorStoreType.CHROMA
    persist_directory: str = None
    collection_name: str = DEFAULT_COLLECTION_NAME
    connection_string: str = None

    class Config:
        arbitrary_types_allowed = True
        extra = "forbid"


class RAGPipelineModel(BaseModel):
    documents: List[Document] = None  # List of documents

    # used for loading an existing vector store
    vector_store_config: VectorStoreConfig = VectorStoreConfig()  # Vector store configuration

    vector_store: VectorStore = vector_store_map[vector_store_config.vector_store_type]  # Vector store
    db_connection_string: str = None  # Database connection string
    table_name: str = DEFAULT_TEST_TABLE_NAME  # table name
    llm: BaseChatModel = None  # Language model
    embeddings_model: Embeddings  # Embeddings model
    rag_prompt_template: str = DEFAULT_RAG_PROMPT_TEMPLATE  # RAG prompt template
    retriever_prompt_template: Union[str, dict] = None  # Retriever prompt template
    retriever_type: RetrieverType = RetrieverType.VECTOR_STORE  # Retriever type

    # Multi retriever specific
    multi_retriever_mode: MultiVectorRetrieverMode = MultiVectorRetrieverMode.BOTH  # Multi retriever mode
    max_concurrency: int = DEFAULT_MAX_CONCURRENCY  # Maximum concurrency
    id_key: int = DEFAULT_ID_KEY  # ID key
    parent_store: BaseStore = None  # Parent store
    text_splitter: TextSplitter = None  # Text splitter
    chunk_size: int = DEFAULT_CHUNK_SIZE  # Chunk size
    chunk_overlap: int = DEFAULT_CHUNK_OVERLAP  # Chunk overlap

    # Auto retriever specific
    auto_retriever_filter_columns: List[str] = None  # Filter columns
    cardinality_threshold: int = DEFAULT_CARDINALITY_THRESHOLD  # Cardinality threshold
    content_column_name: str = DEFAULT_CONTENT_COLUMN_NAME  # content column name (the column we will get embeddings)
    dataset_description: str = DEFAULT_DATASET_DESCRIPTION  # Description of the dataset

    class Config:
        arbitrary_types_allowed = True
        extra = "forbid"

    @classmethod
    def get_field_names(cls):
        return list(cls.model_fields.keys())



from collections import OrderedDict

from mindsdb.integrations.libs.const import HANDLER_CONNECTION_ARG_TYPE as ARG_TYPE


connection_args = OrderedDict(
    code={
        'type': ARG_TYPE.PATH,
        'description': 'The path to model code'
    },
    modules={
        'type': ARG_TYPE.PATH,
        'description': 'The path to model requirements'
    }
)



import time
import requests
from typing import Optional, Dict, Union, List

from mindsdb.utilities import log

logger = log.getLogger(__name__)


class MSGraphAPIBaseClient:
    MICROSOFT_GRAPH_BASE_API_URL: str = "https://graph.microsoft.com/"
    MICROSOFT_GRAPH_API_VERSION: str = "v1.0"
    PAGINATION_COUNT: Optional[int] = 20

    def __init__(self, access_token: str) -> None:
        self.access_token = access_token
        self._group_ids = None

    def _get_api_url(self, endpoint: str) -> str:
        api_url = f"{self.MICROSOFT_GRAPH_BASE_API_URL}{self.MICROSOFT_GRAPH_API_VERSION}/{endpoint}/"
        return api_url

    def _make_request(self, api_url: str, params: Optional[Dict] = None, data: Optional[Dict] = None, method: str = "GET") -> Union[Dict, object]:
        headers = {"Authorization": f"Bearer {self.access_token}"}
        if method == "GET":
            response = requests.get(api_url, headers=headers, params=params)
        elif method == "POST":
            response = requests.post(api_url, headers=headers, json=data)
        else:
            raise NotImplementedError(f"Method {method} not implemented")
        if response.status_code == 429:
            if "Retry-After" in response.headers:
                pause_time = float(response.headers["Retry-After"])
                time.sleep(pause_time)
                response = requests.get(api_url, headers=headers, params=params)
        if response.status_code not in [200, 201]:
            raise requests.exceptions.RequestException(response.text)
        if response.headers["Content-Type"] == "application/octet-stream":
            raw_response = response.content
        else:
            raw_response = response.json()
        return raw_response

    def _get_request_params(self, params: Optional[Dict] = None, pagination: bool = True) -> Dict:
        if self.PAGINATION_COUNT and pagination:
            params = params if params else {}
            if "$top" not in params:
                params["$top"] = self.PAGINATION_COUNT
        return params

    @staticmethod
    def _get_response_value_unsafe(raw_response: Dict) -> List:
        value = raw_response["value"]
        return value

    def _fetch_data(self, endpoint: str, params: Optional[Dict] = None, pagination: bool = True):
        api_url = self._get_api_url(endpoint)
        params = self._get_request_params(params, pagination)
        while api_url:
            raw_response = self._make_request(api_url, params)
            value = self._get_response_value_unsafe(raw_response)
            params = None
            api_url = raw_response.get("@odata.nextLink", "")
            yield value

    def get_user_profile(self):
        api_url = self._get_api_url("me")
        user_profile = self._make_request(api_url)
        return user_profile

    def check_connection(self):
        try:
            self.get_user_profile()
            return True
        except Exception as e:
            logger.error(f'Error connecting to the Microsoft Graph API: {e}!')
            return False



from typing import Optional
import dill
import pandas as pd
from mindsdb.integrations.libs.base import BaseMLEngine
from typing import Dict, Optional
from type_infer.api import infer_types
from tpot import TPOTClassifier, TPOTRegressor
from sklearn.preprocessing import LabelEncoder


class TPOTHandler(BaseMLEngine):
    name = "TPOT"
    def create(self, target: str, df: Optional[pd.DataFrame] = None, args: Optional[Dict] = None) -> None:
        if args is None:
            args = {}
        type_of_cols=infer_types(df,0).dtypes
        target_dtype=type_of_cols[target]


        if target_dtype in ['binary','categorical','tags']:
            model = TPOTClassifier(generations=args.get('generations', 10),
                                       population_size=args.get('population_size', 100),
                                       verbosity=0,
                                       max_time_mins=args.get('max_time_mins', None),
                                       n_jobs=args.get('n_jobs', -1))
            

        elif target_dtype in ['integer','float','quantity'] :
            model = TPOTRegressor(generations=args.get('generations', 10),
                                      population_size=args.get('population_size', 100),
                                      verbosity=0,
                                      max_time_mins=args.get('max_time_mins', None),
                                      n_jobs=args.get('n_jobs', -1))
        

        if df is not None:
            # Separate out the categorical and non-categorical columns
            categorical_cols=[col for col, type_col in type_of_cols.items() if type_col in ('categorical', 'binary')]

            # Fit a LabelEncoder for each categorical column and store it in a dictionary
            le_dict = {}
            for col in categorical_cols:
                le = LabelEncoder()
                le.fit(df[col])
                le_dict[col] = le

                # Encode the categorical column using the fitted LabelEncoder
                df[col] = le.transform(df[col])


            model.fit(df.drop(columns=[target]), df[target])
            self.model_storage.json_set('args', args)
            self.model_storage.file_set('le_dict', dill.dumps(le_dict))
            self.model_storage.file_set('model', dill.dumps(model.fitted_pipeline_))
        else :
            raise Exception(
                "Data is empty!!"
            )
        

    def predict(self, df: pd.DataFrame, args: Optional[Dict] = None) -> pd.DataFrame:

        model=dill.loads(self.model_storage.file_get("model"))
        le_dict=dill.loads(self.model_storage.file_get("le_dict"))
        target=self.model_storage.json_get('args').get("target")

         # Encode the categorical columns in the input DataFrame using the saved LabelEncoders
        for col, le in le_dict.items():
            if col in df.columns:
                df[col] = le.transform(df[col])
                
        # Make predictions using the trained TPOT model
        results=pd.DataFrame(model.predict(df),columns=[target])

        # Decode the predicted categorical values back into their original values
        for col, le in le_dict.items():
            if col in results.columns:
                results[col] = le.inverse_transform(results[col])

        return results
